# type: ignore
# flake8: noqa

from ctypes import sizeof

_IOC_NRBITS = 8
_IOC_TYPEBITS = 8

_IOC_SIZEBITS = 14
_IOC_DIRBITS = 2

_IOC_NRMASK = (1 << _IOC_NRBITS) - 1
_IOC_TYPEMASK = (1 << _IOC_TYPEBITS) - 1
_IOC_SIZEMASK = (1 << _IOC_SIZEBITS) - 1
_IOC_DIRMASK = (1 << _IOC_DIRBITS) - 1

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS

_IOC_NONE = 0
_IOC_WRITE = 1
_IOC_READ = 2


def _IOC(dir, type, nr, size):
    return (dir << _IOC_DIRSHIFT) | (ord(type) << _IOC_TYPESHIFT) | (nr << _IOC_NRSHIFT) | (size << _IOC_SIZESHIFT)


def _IO(type, nr):
    return _IOC(_IOC_NONE, type, nr, 0)


def _IOR(type, nr, size):
    return _IOC(_IOC_READ, type, nr, sizeof(size))


def _IOW(type, nr, size):
    return _IOC(_IOC_WRITE, type, nr, sizeof(size))


def _IOWR(type, nr, size):
    return _IOC(_IOC_READ | _IOC_WRITE, type, nr, sizeof(size))



import json
import mmap
import select
from ctypes import POINTER, c_int16, c_void_p, cast, pointer, sizeof
from dataclasses import dataclass
from enum import Enum, auto
from math import floor
from os import path
from typing import Any, Dict, List, Optional, Tuple, Union

from actfw_core.autofocus import AutoFocuserBase
from actfw_core.capture import Frame
from actfw_core.linux.dma_heap import DMAHeap  # type: ignore
from actfw_core.task import Producer
from actfw_core.v4l2.types import (
    AWB_REGIONS,
    CONTRAST_NUM_POINTS,
    NUM_HISTOGRAM_BINS,
    bcm2835_isp_black_level,
    bcm2835_isp_gain_format,
    bcm2835_isp_gamma,
    bcm2835_isp_lens_shading,
    bcm2835_isp_stats,
    v4l2_ext_control,
)
from actfw_core.v4l2.video import (  # type: ignore
    MEDIA_BUS_FMT,
    V4L2_BUF_TYPE,
    V4L2_CID,
    V4L2_MEMORY,
    V4L2_META_FMT,
    V4L2_PIX_FMT,
    RawVideo,
    V4LConverter,
)

from .util.pad import _PadBase, _PadDiscardingOld

_EMPTY_LIST: List[str] = []

AGC_INTERVAL: int = 3
# TODO: support other than imx219
# pick from https://github.com/kbingham/libcamera/blob/22ffeae04de2e7ce6b2476a35233c790beafb67f/src/ipa/raspberrypi/data/imx219.json#L132-L142 # noqa: E501, B950

MAX_LS_GRID_SIZE = 0x8000
LS_TABLE_W = 16
LS_TABLE_H = 12


# correspond to [libcamera's CameraMode](https://github.com/raspberrypi/libcamera/blob/3fad116f89e0d3497567043cbf6d8c49f1c102db/src/ipa/raspberrypi/controller/camera_mode.h#L19) # noqa: E501, B950
@dataclass(frozen=True, eq=True)
class CameraMode:
    size: Tuple[int, int]
    scale: Tuple[float, float]
    crop: Tuple[int, int]


V1_SENSOR_SIZE = (2592, 1944)
V1_UNICAM_MODES: List[CameraMode] = [
    CameraMode(size=(2592, 1944), scale=(1.0, 1.0), crop=(0, 0)),
    CameraMode(size=(1920, 1080), scale=(1.0, 1.0), crop=(336, 432)),
    CameraMode(size=(1296, 972), scale=(2.0, 2.0), crop=(0, 0)),
    CameraMode(size=(640, 480), scale=(4.0, 4.0), crop=(0, 0)),
]

V2_SENSOR_SIZE = (3280, 2464)
V2_UNICAM_MODES: List[CameraMode] = [
    CameraMode(size=(3280, 2464), scale=(1.0, 1.0), crop=(0, 0)),
    CameraMode(size=(1920, 1080), scale=(1.0, 1.0), crop=(680, 692)),
    CameraMode(size=(1640, 1232), scale=(2.0, 2.0), crop=(0, 0)),
    CameraMode(size=(640, 480), scale=(2.0, 2.0), crop=(1000, 752)),
]

V3_SENSOR_SIZE = (4608, 2592)
V3_UNICAM_MODES: List[CameraMode] = [
    CameraMode(size=(4608, 2592), scale=(1.0, 1.0), crop=(0, 0)),
    CameraMode(size=(2304, 1296), scale=(2.0, 2.0), crop=(0, 0)),
    CameraMode(size=(1536, 864), scale=(2.0, 2.0), crop=(768, 432)),
]

SENSOR_SIZE_MAP = {"ov5647": V1_SENSOR_SIZE, "imx219": V2_SENSOR_SIZE, "imx708": V3_SENSOR_SIZE}


# Used when set values are left to automatic control
class Auto(Enum):
    AUTO = auto()


@dataclass(init=True)
class _DeviceStatus:
    # - update by awb
    gain_r: float = 1.6
    gain_b: float = 1.6
    # - update by agc
    gain: float = 1.0
    shutter_speed: float = 1000.0
    # some device status cache (set by set_unicam_fps)
    vblank: int = 0
    hblank: int = 0
    pixel_late: int = 0
    # vflip & hflip
    vflip: bool = False
    hflip: bool = False


class UnicamIspCapture(Producer[Frame[bytes]]):
    def __init__(
        self,
        unicam: str = "/dev/video0",
        unicam_meta: str = "/dev/video1",
        unicam_subdev: str = "/dev/v4l-subdev0",
        unicam_subdev_meta: str = "/dev/v4l-subdev1",
        isp_in: str = "/dev/video13",
        isp_out_high: str = "/dev/video14",
        isp_out_metadata: str = "/dev/video16",
        size: Tuple[int, int] = (640, 480),
        unicam_size: Optional[Tuple[int, int]] = None,
        crop_size: Optional[Tuple[int, int, int, int]] = None,
        framerate: int = 30,
        expected_format: V4L2_PIX_FMT = V4L2_PIX_FMT.RGB24,
        auto_whitebalance: bool = True,
        init_controls: List[str] = _EMPTY_LIST,
        agc: bool = True,
        target_Y: float = 0.16,  # Temporary set for the developement of agc algorithm
        alsc: bool = True,
        default_color_temperature: int = 4500,
        contrast: bool = True,
        config: Optional[Dict[str, Any]] = None,
        vflip: bool = False,
        hflip: bool = False,
        shutter_time: Union[float, Auto] = Auto.AUTO,
        analogue_gain: Union[float, Auto] = Auto.AUTO,
        auto_focuser: Optional[AutoFocuserBase] = None,
    ) -> None:
        super().__init__()

        self.dma_buffer_num = 4
        self.unicam_meta_buffer_num = 12
        self.isp_out_buffer_num = 4
        self.isp_out_metadata_buffer_num = 2
        self.shared_dma_fds: List[int] = []
        self.sensor_name = self.__get_sensor_name(unicam_subdev)
        self.auto_focuser = auto_focuser
        if self.sensor_name not in ["imx708", "imx219", "ov5647"]:
            raise RuntimeError(f"not supported sensor: {self.sensor_name}")
        self.unicam = RawVideo(
            unicam,
            v4l2_buf_type=V4L2_BUF_TYPE.VIDEO_CAPTURE,
            init_controls=init_controls,
        )
        self.unicam_meta = (
            RawVideo(
                unicam_meta,
                v4l2_buf_type=V4L2_BUF_TYPE.META_CAPTURE,
                init_controls=init_controls,
            )
            if self.auto_focuser is not None
            else None
        )
        self.unicam_meta_bufidx = -1
        self.unicam_subdev = RawVideo(
            unicam_subdev,
            v4l2_buf_type=V4L2_BUF_TYPE.VIDEO_CAPTURE,
            init_controls=init_controls,
        )
        self.unicam_subdev_meta = (
            RawVideo(
                unicam_subdev_meta,
                v4l2_buf_type=V4L2_BUF_TYPE.META_CAPTURE,
                init_controls=init_controls,
            )
            if self.auto_focuser is not None
            else None
        )
        self.isp_in = RawVideo(
            isp_in,
            v4l2_buf_type=V4L2_BUF_TYPE.VIDEO_OUTPUT,
            init_controls=init_controls,
        )
        self.isp_out_high = RawVideo(
            isp_out_high,
            v4l2_buf_type=V4L2_BUF_TYPE.VIDEO_CAPTURE,
            init_controls=init_controls,
        )
        self.isp_out_metadata = RawVideo(
            isp_out_metadata,
            v4l2_buf_type=V4L2_BUF_TYPE.META_CAPTURE,
            init_controls=init_controls,
        )
        self.do_awb = auto_whitebalance
        self.do_agc = agc
        self.do_contrast = contrast
        self.do_alsc = alsc

        (self.expected_width, self.expected_height) = size
        self.expected_pix_format = expected_format
        self.expected_fps = framerate

        self.shutter_time = shutter_time
        self.analogue_gain = analogue_gain

        if not agc and (shutter_time == Auto.AUTO or analogue_gain == Auto.AUTO):
            raise RuntimeError("shutter_time and analogue_gain cannot be AUTO when agc is disabled")

        if unicam_size is not None and crop_size is not None:
            matched_modes = list(filter(lambda mode: mode.size == unicam_size, V2_UNICAM_MODES))
            if len(matched_modes) == 0:
                raise RuntimeError(f"({unicam_size}) is not supported for unicam size")
            self.camera_mode = matched_modes[0]
            self.crop_size = crop_size
        elif unicam_size is None and crop_size is None:
            # Auto selection of unicam_size and crop_size.
            if self.sensor_name == "imx219":
                self.__auto_size_selection_for_imx219()
            elif self.sensor_name == "imx708":
                self.__auto_size_selection_for_imx708()
            else:
                self.__auto_size_selection_for_ov5647()
        else:
            raise RuntimeError("Both unicam_size and crop_size must be None or tuples.")

        # control values
        if self.sensor_name == "imx708":
            config_file = self.__get_subdevice_name(unicam_subdev) + ".json"
        else:
            config_file = self.sensor_name + ".json"
        # config precedence: default < sensor specific < user given
        with open(path.join(path.dirname(__file__), "data", config_file), "r") as f:
            sensor_config = json.load(f)
        sensor_config.update(config or {})

        # black level config
        _bl = sensor_config.get("rpi.black_level", {})
        self.black_level: int = _bl.get("black_level", 4096)

        # lux config
        _lx = sensor_config.get("rpi.lux", {})
        self.reference_shutter_speed: float = _lx.get("reference_shutter_speed", 27685.0)
        self.reference_gain: float = _lx.get("reference_gain", 1.0)
        self.reference_aperture: float = _lx.get("reference_aperture", 1.0)
        self.reference_lux: float = _lx.get("reference_lux", 998.0)
        self.reference_Y: float = _lx.get("reference_Y", 12744.0)
        self.aperture: float = 1.0
        # - update by agc
        _ag = sensor_config.get("rpi.agc", {})
        _ag = _ag["channels"][0] if "channels" in _ag else _ag
        self.shutters = _ag.get("exposure_modes", {}).get("normal", {}).get("shutter")
        self.gains = _ag.get("exposure_modes", {}).get("normal", {}).get("gain")
        self.device_status = _DeviceStatus(vflip=vflip, hflip=hflip)
        self.exposure: float = 100  # `shutter speed(us)` * `analogue gain`
        self.degital_gain: float = 1.0  # Currently, this value is constant.
        self.agc_interval_count: int = 0
        self.target_Y: float = target_Y
        # - update by contrast
        _cr = sensor_config.get("rpi.contrast", {})
        self.ce_enable = _cr.get("ce_enable", True)
        self.brightness: float = _cr.get("brightness", 0.0)
        self.contrast: float = _cr.get("contrast", 1.0)
        self.lo_histogram: float = _cr.get("hi_histogram", 0.01)
        self.lo_level: float = _cr.get("lo_level", 0.015)
        self.lo_max: int = _cr.get("lo_max", 500)
        self.hi_histogram: float = _cr.get("hi_histogram", 0.95)
        self.hi_level: float = _cr.get("hi_level", 0.95)
        self.hi_max: int = _cr.get("hi_max", 0.95)
        gamma_curve = _cr["gamma_curve"]
        self.gamma_curve: List[Tuple[float, float]] = [
            (gamma_curve[2 * n], gamma_curve[2 * n + 1]) for n in range(0, len(gamma_curve) // 2)
        ]

        self.color_temperature = default_color_temperature

        # - update by alsc
        self.ls_table_dma_heap_fd = DMAHeap("/dev/dma_heap/linux,cma").alloc("_ls_grid", MAX_LS_GRID_SIZE)
        self.ls_table_mm = mmap.mmap(
            self.ls_table_dma_heap_fd,
            MAX_LS_GRID_SIZE,
            flags=mmap.MAP_SHARED,
            prot=mmap.PROT_READ | mmap.PROT_WRITE,
        )
        _alsc = sensor_config.get("rpi.alsc", {})
        self.calibrations_Cr: List[Dict[str, Any]] = _alsc["calibrations_Cr"]
        self.calibrations_Cb: List[Dict[str, Any]] = _alsc["calibrations_Cb"]
        self.luminace_lut: List[float] = _alsc["luminance_lut"]
        self.luminace_strength: float = _alsc["luminance_strength"]

        # setup
        self.converter = V4LConverter(self.isp_out_high.device_fd)
        self.__setup_pipeline()
        # autofocus
        if self.auto_focuser is not None:
            self.auto_focuser.set_unicam_config(
                sensor_config,
                unicam_width=self.unicam_width,
                sensor_size=SENSOR_SIZE_MAP[self.sensor_name],
                post_crop_size=self.crop_size,
                camera_mode_scale=self.camera_mode.scale,
                camera_mode_crop_size=self.camera_mode.crop,
                callback_fn=self.__set_focus_absolute,
            )
        self.output_fmt = self.converter.try_convert(
            self.isp_out_high.fmt,
            self.expected_width,
            self.expected_height,
            self.expected_pix_format,
        )

        self.__request_buffer()

    def set_exposure_settings(self, shutter_time: Union[float, Auto], analogue_gain: Union[float, Auto]) -> None:
        """Set shutter_time and analogue_gain.

        Args:
            shutter_time (float or Auto): [μsec] (e.g. 60000). If set to Auto, shutter_time is set with agc algorithm .
            analogue_gain (float or Auto): (>=0) (e.g. 2.5). If set to Auto, analogue_gain is set with agc algorithm .
        """
        if not self.do_agc and (shutter_time == Auto.AUTO or analogue_gain == Auto.AUTO):
            raise RuntimeError("shutter_time and analogue_gain cannot be AUTO when agc is disabled")
        self.shutter_time = shutter_time
        self.analogue_gain = analogue_gain

    def set_exposure_time(self, ms: Optional[int] = None) -> bool:
        """Set exposure time.

        This function is no longer supported. Use set_exposure_settings instead.

        Returns:
            False
        """
        return False

    def __get_sensor_name(self, subdev: str) -> str:
        with open(f"/sys/class/video4linux/{subdev[5:]}/device/name") as f:
            sensor_name = f.read()
        return sensor_name.rstrip()

    def __get_subdevice_name(self, subdev: str) -> str:
        with open(f"/sys/class/video4linux/{subdev[5:]}/device/video4linux/{subdev[5:]}/name") as f:
            subdevice_name = f.read()
        return subdevice_name.rstrip()

    def __setup_pipeline(self) -> None:
        # setup unicam
        if self.sensor_name in ["imx219", "imx708"]:
            # imx219 is flipped by default
            if not (
                self.unicam_subdev.set_vertical_flip(not self.device_status.vflip)
                and self.unicam_subdev.set_horizontal_flip(not self.device_status.hflip)
            ):
                raise RuntimeError("fail to setup unicam subdevice node")
            # https://www.kernel.org/doc/html/v5.15/userspace-api/media/v4l/subdev-formats.html?highlight=media_bus_fmt
            if self.device_status.vflip and self.device_status.hflip:
                bus_fmt = MEDIA_BUS_FMT.SRGGB10_1X10
                self.unicam_format = V4L2_PIX_FMT.SRGGB10P
            elif self.device_status.vflip and not self.device_status.hflip:
                bus_fmt = MEDIA_BUS_FMT.SGRBG10_1X10
                self.unicam_format = V4L2_PIX_FMT.SGRBG10P
            elif not self.device_status.vflip and self.device_status.hflip:
                bus_fmt = MEDIA_BUS_FMT.SGBRG10_1X10
                self.unicam_format = V4L2_PIX_FMT.SGBRG10P
            else:
                bus_fmt = MEDIA_BUS_FMT.SBGGR10_1X10
                self.unicam_format = V4L2_PIX_FMT.SBGGR10P
        else:
            if not (
                self.unicam_subdev.set_vertical_flip(self.device_status.vflip)
                and self.unicam_subdev.set_horizontal_flip(self.device_status.hflip)
            ):
                raise RuntimeError("fail to setup unicam subdevice node")
            if self.device_status.vflip and self.device_status.hflip:
                bus_fmt = MEDIA_BUS_FMT.SGRBG10_1X10
                self.unicam_format = V4L2_PIX_FMT.SGRBG10P
            elif self.device_status.vflip and not self.device_status.hflip:
                bus_fmt = MEDIA_BUS_FMT.SRGGB10_1X10
                self.unicam_format = V4L2_PIX_FMT.SRGGB10P
            elif not self.device_status.vflip and self.device_status.hflip:
                bus_fmt = MEDIA_BUS_FMT.SBGGR10_1X10
                self.unicam_format = V4L2_PIX_FMT.SBGGR10P
            else:
                bus_fmt = MEDIA_BUS_FMT.SGBRG10_1X10
                self.unicam_format = V4L2_PIX_FMT.SGBRG10P

        self.unicam_subdev.set_subdev_format(*self.camera_mode.size, bus_fmt)
        if (
            self.unicam_subdev.subdev_fmt.format.width != self.camera_mode.size[0]
            or self.unicam_subdev.subdev_fmt.format.height != self.camera_mode.size[1]
        ):
            raise RuntimeError("fail to setup unicam device node")

        self.unicam_width = self.unicam_subdev.subdev_fmt.format.width
        self.unicam_height = self.unicam_subdev.subdev_fmt.format.height
        (unicam_width, unicam_height, unicam_format) = self.unicam.set_pix_format(
            self.unicam_width, self.unicam_height, self.unicam_format
        )
        if unicam_width != self.unicam_width or unicam_height != self.unicam_height or unicam_format != self.unicam_format:
            raise RuntimeError("fail to setup unicam device node")

        self.__set_unicam_fps()
        if self.do_agc:
            init_shutter_time = self.shutters[len(self.shutters) // 2]  # (us)
            init_analogue_gain = self.gains[len(self.gains) // 2]
            self.__set_unicam_exposure(init_analogue_gain, init_shutter_time)
            self.exposure = init_shutter_time * init_analogue_gain

        # current alsc implementation does not change lens_shading table dynamically
        if self.do_alsc:
            self.__alsc()

        # sutup isp_in
        bl = bcm2835_isp_black_level()
        bl.enabled = 1
        bl.black_level_r = self.black_level
        bl.black_level_g = self.black_level
        bl.black_level_b = self.black_level
        black_level = v4l2_ext_control()
        black_level.id = V4L2_CID.USER_BCM2835_ISP_BLACK_LEVEL
        black_level.size = sizeof(bcm2835_isp_black_level)
        black_level.ptr = cast(pointer(bl), c_void_p)
        self.isp_in.set_ext_controls([black_level])
        (isp_in_width, isp_in_height, isp_in_format) = self.isp_in.set_pix_format(
            self.unicam_width, self.unicam_height, self.unicam_format
        )
        if isp_in_width != self.unicam_width or isp_in_height != self.unicam_height or isp_in_format != self.unicam_format:
            raise RuntimeError("fail to setup isp input")

        if self.crop_size is not None:
            self.isp_in.set_selection(*self.crop_size)

        # setup isp_out_high

        # TODO: Ensure that the value of bytesperline is equal to width or handle padding appropriately.
        # One possible solution is to use only width of multiples of 32.
        (
            isp_out_width,
            isp_out_height,
            isp_out_format,
        ) = self.isp_out_high.set_pix_format(self.expected_width, self.expected_height, self.expected_pix_format)

        self.output_fmt = self.converter.try_convert(
            self.isp_out_high.fmt,
            self.expected_width,
            self.expected_height,
            self.expected_pix_format,
        )

        # setup isp_out_metadata
        self.isp_out_metadata.set_meta_format(V4L2_META_FMT.BCM2835_ISP_STATS, sizeof(bcm2835_isp_stats))

    def __set_unicam_fps(self) -> None:
        ctrls = self.unicam_subdev.get_ext_controls([V4L2_CID.HBLANK, V4L2_CID.PIXEL_RATE])
        hblank = ctrls[0].value
        pixel_late = ctrls[1].value64

        expected_pixel_per_second = pixel_late // self.expected_fps
        expected_line_num = expected_pixel_per_second // (self.unicam_width + hblank)
        expected_vblank = expected_line_num - self.unicam_height
        # TODO: check supported vblank range
        vblank_ctrl = v4l2_ext_control()
        vblank_ctrl.id = V4L2_CID.VBLANK
        vblank_ctrl.value = expected_vblank
        ctrls = self.unicam_subdev.set_ext_controls([vblank_ctrl])

        self.device_status.hblank = hblank
        self.device_status.vblank = expected_vblank
        self.device_status.pixel_late = pixel_late

    def __set_unicam_exposure(self, analogue_gain: float, shutter_time: float) -> None:
        # convert analogue_gain to V4L2_CID.ANALOGUE_GAIN
        unicam_subdev_analogue_gain = self.__v4l2_control_value_for_analogue_gain(analogue_gain)

        # convert shutter time to V4L2_CID.EXPOSURE
        time_per_line = (self.unicam_width + self.device_status.hblank) * (1.0 / self.device_status.pixel_late) * 1e6
        unicam_subdev_exposure = shutter_time / time_per_line

        exposure_ctrl = v4l2_ext_control()
        exposure_ctrl.id = V4L2_CID.EXPOSURE
        exposure_ctrl.value = int(unicam_subdev_exposure)
        gain_ctrl = v4l2_ext_control()
        gain_ctrl.id = V4L2_CID.ANALOGUE_GAIN
        gain_ctrl.value = int(unicam_subdev_analogue_gain)
        self.unicam_subdev.set_ext_controls([exposure_ctrl, gain_ctrl])

    def __set_focus_absolute(self, focus_val: int) -> None:
        assert self.unicam_subdev_meta is not None
        focus_control = v4l2_ext_control()
        focus_control.id = V4L2_CID.FOCUS_ABSOLUTE
        focus_control.value = focus_val
        self.unicam_subdev_meta.set_ext_controls([focus_control])

    def __v4l2_control_value_for_analogue_gain(self, analogue_gain: float) -> float:
        if self.sensor_name == "imx219":
            return 256 - (
                256 / analogue_gain
            )  # https://github.com/raspberrypi/libcamera/blob/3fad116f89e0d3497567043cbf6d8c49f1c102db/src/ipa/raspberrypi/cam_helper_imx219.cpp#L67 # noqa: E501, B950
        elif self.sensor_name == "imx708":
            return 1024 - (
                1024 / analogue_gain
            )  # https://github.com/raspberrypi/libcamera/blob/6ddd79b5bdbedc1f61007aed35391f1559f9e29a/src/ipa/rpi/cam_helper/cam_helper_imx708.cpp#L103-L106 # noqa: E501, B950
        elif self.sensor_name == "ov5647":
            return (
                analogue_gain * 16.0
            )  # https://github.com/raspberrypi/libcamera/blob/3fad116f89e0d3497567043cbf6d8c49f1c102db/src/ipa/raspberrypi/cam_helper_ov5647.cpp#L45 # noqa: E501, B950
        else:
            raise RuntimeError(f"not supported sensor: {self.sensor_name}")

    def __calc_crop_size(
        self,
        isp_out_width: int,
        isp_out_height: int,
        unicam_width: int,
        unicam_height: int,
    ) -> Tuple[int, int, int, int]:
        scale = min(unicam_height / isp_out_height, unicam_width / isp_out_width)
        w = round(isp_out_width * scale)
        h = round(isp_out_height * scale)
        assert w <= unicam_width and h <= unicam_height

        left = int((unicam_width - w) / 2)
        top = int((unicam_height - h) / 2)
        return (left, top, w, h)

    def __auto_size_selection_for_imx708(self) -> None:
        if self.expected_width <= 1920 and self.expected_height <= 1080:
            if self.expected_fps <= 50:
                self.camera_mode = V3_UNICAM_MODES[1]
            else:
                self.camera_mode = V3_UNICAM_MODES[2]
        else:
            self.camera_mode = V3_UNICAM_MODES[0]
        self.crop_size = self.__calc_crop_size(self.expected_width, self.expected_height, *self.camera_mode.size)

    def __auto_size_selection_for_imx219(self) -> None:
        # Support only v2 camera module.
        if self.expected_width <= 1280 and self.expected_height <= 720:
            if self.expected_fps <= 40:
                self.camera_mode = V2_UNICAM_MODES[2]
                self.crop_size = self.__calc_crop_size(self.expected_width, self.expected_height, *self.camera_mode.size)
            else:
                if abs(self.expected_width / self.expected_height - 16 / 9) < 0.05:
                    self.camera_mode = V2_UNICAM_MODES[2]
                    self.crop_size = (180, 256, 1280, 720)
                else:
                    self.camera_mode = V2_UNICAM_MODES[3]
                    self.crop_size = self.__calc_crop_size(
                        self.expected_width,
                        self.expected_height,
                        *self.camera_mode.size,
                    )
        else:
            self.camera_mode = V2_UNICAM_MODES[0]
            self.crop_size = self.__calc_crop_size(self.expected_width, self.expected_height, *self.camera_mode.size)

    def __auto_size_selection_for_ov5647(self) -> None:
        # TODO: fix to make compatiblity with buster
        self.camera_mode = V1_UNICAM_MODES[0]
        self.crop_size = self.__calc_crop_size(self.expected_width, self.expected_height, *self.camera_mode.size)

    def capture_size(self) -> Tuple[int, int]:
        return (self.output_fmt.fmt.pix.width, self.output_fmt.fmt.pix.height)

    def __request_buffer(self) -> None:
        self.dma_buffer_num = self.unicam.request_buffers(self.dma_buffer_num, V4L2_MEMORY.MMAP)
        self.dma_fds = self.unicam.export_buffers()
        if self.unicam.request_buffers(self.dma_buffer_num, V4L2_MEMORY.DMABUF, self.dma_fds) != self.dma_buffer_num:
            raise RuntimeError("dma buffer count mismatch with the number of shared dma file descrptors")
        if self.isp_in.request_buffers(self.dma_buffer_num, V4L2_MEMORY.DMABUF, self.dma_fds) != self.dma_buffer_num:
            raise RuntimeError("dma buffer count mismatch with the number of shared dma file descrptors")
        self.isp_out_buffer_num = self.isp_out_high.request_buffers(self.isp_out_buffer_num, V4L2_MEMORY.MMAP)
        self.isp_out_metadata_buffer_num = self.isp_out_metadata.request_buffers(
            self.isp_out_metadata_buffer_num, V4L2_MEMORY.MMAP
        )
        if self.unicam_meta is not None:
            self.unicam_meta_buffer_num = self.unicam_meta.request_buffers(self.unicam_meta_buffer_num, V4L2_MEMORY.MMAP)

    def __unicam2isp(self) -> None:
        buffer = self.unicam.dequeue_buffer_nonblocking(v4l2_memory=V4L2_MEMORY.DMABUF)
        if buffer is None:
            return
        self.isp_in.queue_buffer(buffer.buf.index)

    def __isp2unicam(self) -> None:
        buffer = self.isp_in.dequeue_buffer_nonblocking(v4l2_memory=V4L2_MEMORY.DMABUF)
        if buffer is None:
            return
        self.unicam.queue_buffer(buffer.buf.index)

    def __receive_unicam_meta(self) -> None:
        if self.unicam_meta is not None:
            meta_buffer = self.unicam_meta.dequeue_buffer_nonblocking(v4l2_memory=V4L2_MEMORY.MMAP)
            self.unicam_meta_bufidx = meta_buffer.buf.index
            if self.auto_focuser is not None:
                self.auto_focuser.parse_pdaf_and_update_focus(meta_buffer)
            self.unicam_meta.queue_buffer(meta_buffer.buf.index)

    def __produce_image_from_isp(self) -> None:
        buffer = self.isp_out_high.dequeue_buffer_nonblocking(v4l2_memory=V4L2_MEMORY.MMAP)
        if buffer is None:
            return

        dst = self.converter.convert(buffer, self.isp_out_high.fmt, self.output_fmt)
        frame = Frame(dst)
        self._outlet(frame)
        self.isp_out_high.queue_buffer(buffer.buf.index)

    def __calc_lux(self, isp_stats: bcm2835_isp_stats) -> None:
        current_aperture = self.aperture
        current_gain = self.device_status.gain
        current_shutter_speed = self.device_status.shutter_speed

        hist_sum = 0
        hist_num = 0
        hist = isp_stats.hist[0].g_hist
        for i in range(NUM_HISTOGRAM_BINS):
            hist_sum += hist[i] * i
            hist_num += hist[i]
        current_Y = float(hist_sum) / float(hist_num) + 0.5
        gain_ratio = self.reference_gain / current_gain
        shutter_speed_ratio = self.reference_shutter_speed / current_shutter_speed
        aperture_ratio = self.reference_aperture / current_aperture
        Y_ratio = current_Y * (65536.0 / NUM_HISTOGRAM_BINS) / self.reference_Y
        estimated_lux = shutter_speed_ratio * gain_ratio * aperture_ratio * aperture_ratio * Y_ratio * self.reference_lux

        self.lux = estimated_lux

    def __get_cal_table(self, ct: float, calibrations: List[Dict[str, Any]]) -> List[float]:
        if len(calibrations) == 0:
            return [1.0] * AWB_REGIONS
        elif ct <= calibrations[0]["ct"]:
            return calibrations[0]["table"].copy()  # type: ignore
        elif calibrations[-1]["ct"] <= ct:
            return calibrations[-1]["table"].copy()  # type: ignore
        else:
            idx = 0
            while calibrations[idx + 1]["ct"] < ct:
                idx += 1

            ct0 = calibrations[idx]["ct"]
            ct1 = calibrations[idx + 1]["ct"]
            calibration0 = calibrations[idx]["table"]
            calibration1 = calibrations[idx + 1]["table"]
            return [(x0 * (ct1 - ct) + x1 * (ct - ct0)) / (ct1 - ct0) for (x0, x1) in zip(calibration0, calibration1)]

    # simplified version of [alsc algorithm in libcamera](https://github.com/raspberrypi/libcamera/blob/3fad116f89e0d3497567043cbf6d8c49f1c102db/src/ipa/raspberrypi/controller/rpi/alsc.cpp#L772) # noqa: E501, B950
    # with the following omitted
    #  - dynamic lens shading table calculation based on estimated colour templature
    #  - "Adaptive ALSC Algorithm"
    def __alsc(self) -> None:
        ct = self.color_temperature

        cal_table_r = self.__get_cal_table(ct, self.calibrations_Cr)
        cal_table_b = self.__get_cal_table(ct, self.calibrations_Cb)

        def normalize(table: List[float]) -> List[float]:
            m = min(table)
            return [x / m for x in table]

        cal_table_r = self.__resample_cal_table(cal_table_r, self.camera_mode)
        cal_table_b = self.__resample_cal_table(cal_table_b, self.camera_mode)
        luminace_table = self.__resample_cal_table(self.luminace_lut, self.camera_mode)

        self.ls_table_r = normalize(
            [r * ((lut - 1) * self.luminace_strength + 1) for (r, lut) in zip(cal_table_r, luminace_table)]
        )
        self.ls_table_g = normalize([1.0 * ((lut - 1) * self.luminace_strength + 1) for lut in luminace_table])
        self.ls_table_b = normalize(
            [b * ((lut - 1) * self.luminace_strength + 1) for (b, lut) in zip(cal_table_b, luminace_table)]
        )

        self.__apply_ls_tables()

    # correspond to [resampleCalTable](https://github.com/raspberrypi/libcamera/blob/3fad116f89e0d3497567043cbf6d8c49f1c102db/src/ipa/raspberrypi/controller/rpi/alsc.cpp#L463) # noqa: E501, B950
    def __resample_cal_table(self, src: List[float], camera_mode: CameraMode) -> List[float]:
        _sensor_size = SENSOR_SIZE_MAP[self.sensor_name]
        assert len(src) == LS_TABLE_W * LS_TABLE_H
        new_table: List[float] = []

        x_lo: List[int] = [0] * LS_TABLE_W
        x_hi: List[int] = [0] * LS_TABLE_W
        xf: List[float] = [0] * LS_TABLE_W
        scale_x = _sensor_size[0] / (camera_mode.size[0] * camera_mode.scale[0])
        offset_x = (camera_mode.crop[0] / _sensor_size[0]) * LS_TABLE_W
        x = (0.5 / scale_x) + offset_x - 0.5
        x_inc = 1 / scale_x
        for i in range(0, LS_TABLE_W):
            x_lo[i] = floor(x)
            xf[i] = x - x_lo[i]
            x_hi[i] = min(x_lo[i] + 1, LS_TABLE_W - 1)
            x_lo[i] = max(x_lo[i], 0)
            x += x_inc

        scale_y = _sensor_size[1] / (camera_mode.size[1] * camera_mode.scale[1])
        offset_y = (camera_mode.crop[1] / _sensor_size[1]) * LS_TABLE_H
        y = (0.5 / scale_y) + offset_y - 0.5
        y_inc = 1 / scale_y
        for _ in range(0, LS_TABLE_H):
            y_lo = floor(y)
            yf = y - y_lo
            y_hi = min(y_lo + 1, LS_TABLE_H - 1)
            y_lo = max(y_lo, 0)
            row_above = src[y_lo * LS_TABLE_W : (y_lo + 1) * LS_TABLE_W]
            row_below = src[y_hi * LS_TABLE_W : (y_hi + 1) * LS_TABLE_W]
            for i in range(0, LS_TABLE_W):
                above = row_above[x_lo[i]] * (1 - xf[i]) + row_above[x_hi[i]] * xf[i]
                below = row_below[x_lo[i]] * (1 - xf[i]) + row_below[x_hi[i]] * xf[i]
                new_table.append(above * (1 - yf) + below * yf)

            y += y_inc

        assert len(new_table) == LS_TABLE_W * LS_TABLE_H
        return new_table

    # correspond to [applyLS](https://github.com/raspberrypi/libcamera/blob/1c4c323e5d684b57898c083ed2f1af313bf6a98d/src/ipa/raspberrypi/raspberrypi.cpp#L1343) # noqa: E501, B950
    def __apply_ls_tables(self) -> None:
        assert len(self.ls_table_b) == LS_TABLE_W * LS_TABLE_H
        assert len(self.ls_table_r) == LS_TABLE_W * LS_TABLE_H
        assert len(self.ls_table_g) == LS_TABLE_W * LS_TABLE_H
        cell_size_candidate = [16, 32, 64, 128, 256]
        for i in range(0, len(cell_size_candidate)):
            cell_size = cell_size_candidate[i]
            w = (self.unicam_width + cell_size - 1) // cell_size
            h = (self.unicam_height + cell_size - 1) // cell_size
            if w < 64 and h <= 48:
                break

        w += 1
        h += 1

        self.__populate_ls_table(self.ls_table_r, w, h, self.ls_table_mm, 0)
        self.__populate_ls_table(self.ls_table_g, w, h, self.ls_table_mm, 1 * (2 * w * h))
        self.ls_table_mm[2 * (2 * w * h) : 3 * (2 * w * h)] = memoryview(self.ls_table_mm)[1 * (2 * w * h) : 2 * (2 * w * h)]
        self.__populate_ls_table(self.ls_table_b, w, h, self.ls_table_mm, 3 * (2 * w * h))

        ls = bcm2835_isp_lens_shading()
        ls.enabled = 1
        ls.grid_cell_size = cell_size
        ls.grid_width = w
        ls.grid_stride = w
        ls.grid_height = h
        ls.dmabuf = self.ls_table_dma_heap_fd
        ls.ref_transform = 0
        ls.corner_sampled = 1
        ls.gain_format = bcm2835_isp_gain_format.GAIN_FORMAT_U4P10

        ls_ctrl = v4l2_ext_control()
        ls_ctrl.id = V4L2_CID.USER_BCM2835_ISP_LENS_SHADING
        ls_ctrl.size = sizeof(bcm2835_isp_lens_shading)
        ls_ctrl.ptr = cast(pointer(ls), c_void_p)
        self.isp_in.set_ext_controls([ls_ctrl])

    # correspond to [resampleTable](https://github.com/raspberrypi/libcamera/blob/1c4c323e5d684b57898c083ed2f1af313bf6a98d/src/ipa/raspberrypi/raspberrypi.cpp#L1403) # noqa: E501, B950
    def __populate_ls_table(self, src: List[float], dst_w: int, dst_h: int, ls_table: mmap.mmap, offset: int) -> None:
        assert len(src) == LS_TABLE_W * LS_TABLE_H
        x_lo: List[int] = [0] * dst_w
        xf: List[float] = [0.0] * dst_w
        x_hi: List[int] = [0] * dst_w

        x = -0.5
        x_inc = LS_TABLE_W / (dst_w - 1)
        for i in range(0, dst_w):
            x_lo[i] = floor(x)
            xf[i] = x - x_lo[i]
            x_hi[i] = min(x_lo[i] + 1, LS_TABLE_W - 1)
            x_lo[i] = max(x_lo[i], 0)
            x += x_inc

        ls_table_idx = offset
        y = -0.5
        y_inc = LS_TABLE_H / (dst_h - 1)
        for _ in range(0, dst_h):
            y_lo = floor(y)
            yf = y - y_lo
            y_hi = min(y_lo + 1, LS_TABLE_H - 1)
            y_lo = max(y_lo, 0)
            row_above = src[y_lo * LS_TABLE_W : (y_lo + 1) * LS_TABLE_W]
            row_below = src[y_hi * LS_TABLE_W : (y_hi + 1) * LS_TABLE_W]
            for i in range(0, dst_w):
                above = row_above[x_lo[i]] * (1 - xf[i]) + row_above[x_hi[i]] * xf[i]
                below = row_below[x_lo[i]] * (1 - xf[i]) + row_below[x_hi[i]] * xf[i]
                result = floor(1024 * (above * (1 - yf) + below * yf) + 0.5)
                result = min(result, 16383)
                result_as_bytes = bytes(c_int16(result))  # type: ignore
                assert len(result_as_bytes) == 2  # 16 bits
                ls_table[ls_table_idx : ls_table_idx + 2] = result_as_bytes
                ls_table_idx += 2

            y += y_inc

        assert ls_table_idx == offset + (2 * dst_w * dst_h)

    def __calculate_y(self, stats: bcm2835_isp_stats, additional_gain: float) -> float:
        PIPELINE_BITS = 13  # https://github.com/kbingham/libcamera/blob/f995ff25a3326db90513d1fa936815653f7cade0/src/ipa/raspberrypi/controller/rpi/agc.cpp#L31 # noqa: E501, B950
        r_sum = 0
        g_sum = 0
        b_sum = 0
        pixel_sum = 0
        for region in stats.agc_stats:
            counted = region.counted
            r_sum += min(region.r_sum * additional_gain, ((1 << PIPELINE_BITS) - 1) * counted)
            b_sum += min(region.b_sum * additional_gain, ((1 << PIPELINE_BITS) - 1) * counted)
            g_sum += min(region.g_sum * additional_gain, ((1 << PIPELINE_BITS) - 1) * counted)
            pixel_sum += counted

        if pixel_sum == 0:
            return 0

        y_sum = r_sum * self.device_status.gain_r * 0.299 + b_sum * self.device_status.gain_b * 0.144 + g_sum * 0.587

        return y_sum / pixel_sum / (1 << PIPELINE_BITS)

    def __agc(self, stats: bcm2835_isp_stats) -> None:
        # compute addtional gain to acheive target_Y
        # https://github.com/raspberrypi/libcamera/blob/1c4c323e5d684b57898c083ed2f1af313bf6a98d/src/ipa/raspberrypi/controller/rpi/agc.cpp#L572-L582 # noqa: E501, B950
        gain = 1.0
        for _ in range(8):
            initial_Y = self.__calculate_y(stats, gain)
            extra_gain = min(10.0, self.target_Y / (initial_Y + 0.001))
            gain *= extra_gain
            if extra_gain < 1.01:
                break

        max_exposure = self.shutters[-1] * self.gains[-1]
        target_exposure = min(self.exposure * gain, max_exposure)

        # decompose target_exposure to analogue_gain and shutter_time
        # TODO: Also decompose to digital gain.
        # analog_gain==AUTOの場合のみ,analog_gainを動かす
        analogue_gain = self.gains[0] if self.analogue_gain == Auto.AUTO else self.analogue_gain
        shutter_time = self.shutters[0] if self.shutter_time == Auto.AUTO else self.shutter_time
        if shutter_time * analogue_gain < target_exposure:
            for stage in range(1, len(self.gains)):
                # fix gain, increase shutter time
                if self.shutter_time == Auto.AUTO:
                    max_shutter_time = self.shutters[stage]
                    if max_shutter_time * analogue_gain >= target_exposure:
                        shutter_time = target_exposure / analogue_gain
                        break
                    shutter_time = max_shutter_time

                # fix shutter time, increase gain
                if self.analogue_gain == Auto.AUTO:
                    max_analogue_gain = self.gains[stage]
                    if shutter_time * max_analogue_gain >= target_exposure:
                        analogue_gain = target_exposure / shutter_time
                        break
                    analogue_gain = max_analogue_gain

        # may need flicker avoidance here. ref. https://github.com/kbingham/libcamera/blob/d7415bc4e46fe8aa25a495c79516d9882a35a5aa/src/ipa/raspberrypi/controller/rpi/agc.cpp#L724 # noqa: E501, B950
        self.device_status.shutter_speed = shutter_time
        self.device_status.gain = analogue_gain  # TODO: 変数名整理 (self.analogu_gain, self.shutter_time との兼ね合い)
        self.exposure = shutter_time * analogue_gain
        self.__set_unicam_exposure(analogue_gain, shutter_time)

    def __awb(self, stats: bcm2835_isp_stats) -> None:
        sum_r = 0
        sum_b = 0
        sum_g = 0
        # len(stats.awb_stats) == 192
        for region in stats.awb_stats:
            sum_r += region.r_sum
            sum_b += region.b_sum
            sum_g += region.g_sum
        self.device_status.gain_r = sum_g / (sum_r + 1)
        self.device_status.gain_b = sum_g / (sum_b + 1)
        red_balance_ctrl = v4l2_ext_control()
        red_balance_ctrl.id = V4L2_CID.RED_BALANCE
        red_balance_ctrl.value64 = int(self.device_status.gain_r * 1000)
        blue_balance_ctrl = v4l2_ext_control()
        blue_balance_ctrl.id = V4L2_CID.BLUE_BALANCE
        blue_balance_ctrl.value64 = int(self.device_status.gain_b * 1000)
        self.isp_in.set_ext_controls([red_balance_ctrl, blue_balance_ctrl])

    # linearly find appropriate range, this might be inefficient
    def __find_span(self, gamma_curve: List[Tuple[float, float]], x: float) -> int:
        last_span = len(gamma_curve) - 2
        for i in range(0, last_span + 1):
            if gamma_curve[i][0] <= x < gamma_curve[i + 1][0]:
                return i
        return last_span

    def __eval_gamma_curve(self, gamma_curve: List[Tuple[float, float]], x: float) -> float:
        span = self.__find_span(gamma_curve, x)
        return gamma_curve[span][1] + (x - gamma_curve[span][0]) * (gamma_curve[span + 1][1] - gamma_curve[span][1]) / (
            gamma_curve[span + 1][0] - gamma_curve[span][0]
        )

    # gamma2(gamma1(x)) with care of boundary values
    def __compose_gamma_curve(
        self,
        one: List[Tuple[float, float]],
        other: List[Tuple[float, float]],
        eps: float = 1e-6,
    ) -> List[Tuple[float, float]]:
        this_x = one[0][0]
        this_y = one[0][1]
        this_span = 0
        other_span = self.__find_span(other, this_y)
        result = [(this_y, self.__eval_gamma_curve(other, this_y))]
        while this_span != len(one) - 1:
            dx = one[this_span + 1][0] - one[this_span][0]
            dy = one[this_span + 1][1] - one[this_span][1]
            if abs(dy) > eps and other_span + 1 < len(other) and one[this_span + 1][1] >= other[other_span + 1][0] + eps:
                # next control point in result will be where this
                # function's y reaches the next span in other
                this_x = one[this_span][0] + (other[other_span + 1][0] - one[this_span][1]) * dx / dy
                other_span += 1
                this_y = other[other_span][0]
            elif abs(dy) > eps and other_span > 0 and one[this_span + 1][1] <= other[other_span - 1][0] - eps:
                this_x = one[this_span][0] + (other[other_span + 1][0] - one[this_span][1]) * dx / dy
                other_span -= 1
                this_y = other[other_span][0]
            else:
                this_span += 1
                this_x = one[this_span][0]
                this_y = one[this_span][1]
            if result[-1][0] + eps < this_x:
                result.append((this_x, self.__eval_gamma_curve(other, this_y)))
        return result

    def __histogram_cumulative(self, histogram: List[int]) -> List[int]:
        cumulative = [0]
        for i in range(0, len(histogram)):
            cumulative.append(cumulative[-1] + histogram[i])
        return cumulative

    def __cumulative_quantile(self, cumulative: List[int], q: float, first: int = -1, last: int = -1) -> float:
        if first == -1:
            first = 0
        if last == -1:
            last = len(cumulative) - 2
        assert first <= last
        items = int(q * cumulative[-1])
        while first < last:
            middle = (first + last) // 2
            if cumulative[middle + 1] > items:
                last = middle
            else:
                first = middle + 1
        assert items >= cumulative[first] and items <= cumulative[last + 1]
        frac = (
            0
            if cumulative[first + 1] == cumulative[first]
            else (items - cumulative[first]) / (cumulative[first + 1] - cumulative[first])
        )
        return first + frac

    def __compute_stretch_curve(self, histogram: List[int]) -> List[Tuple[float, float]]:
        enhance = [(0.0, 0.0)]
        eps = 1e-6

        # If the start of the histogram is rather empty, try to pull it down a
        # bit.
        cumulative = self.__histogram_cumulative(histogram)
        hist_lo = self.__cumulative_quantile(cumulative, self.lo_histogram) * (65536 / NUM_HISTOGRAM_BINS)
        level_lo = self.lo_level * 65536
        hist_lo = max(level_lo, min(65535, min(hist_lo, level_lo + self.lo_max)))
        if enhance[-1][0] + eps < hist_lo:
            enhance.append((hist_lo, level_lo))

        # Keep the mid-point (median) in the same place, though, to limit the
        # apparent amount of global brightness shift.
        mid = self.__cumulative_quantile(cumulative, 0.5) * (65536 / NUM_HISTOGRAM_BINS)
        if enhance[-1][0] + eps < mid:
            enhance.append((mid, mid))

        # If the top to the histogram is empty, try to pull the pixel values
        # there up.
        hist_hi = self.__cumulative_quantile(cumulative, self.hi_histogram) * (65536 / NUM_HISTOGRAM_BINS)
        level_hi = self.hi_level * 65536
        hist_hi = min(level_hi, max(0.0, max(hist_hi, level_hi - self.hi_max)))
        if enhance[-1][0] + eps < hist_hi:
            enhance.append((hist_hi, level_hi))
        if enhance[-1][0] + eps < 65535:
            enhance.append((65535, 65535))
        return enhance

    def __fill_in_contrast_status(self, gm: bcm2835_isp_gamma, gamma_curve: List[Tuple[float, float]]) -> None:
        gm.enabled = 1
        for i in range(0, CONTRAST_NUM_POINTS - 1):
            if i < 16:
                x = i * 1024
            elif i < 24:
                x = (i - 16) * 2048 + 16384
            else:
                x = (i - 24) * 4096 + 32768
            gm.x[i] = x
            gm.y[i] = int(min(65535.0, self.__eval_gamma_curve(gamma_curve, x)))

        gm.x[CONTRAST_NUM_POINTS - 1] = 65535
        gm.y[CONTRAST_NUM_POINTS - 1] = 65535

    def __contrast_control(self, isp_stats: bcm2835_isp_stats) -> None:
        histogram = isp_stats.hist[0].g_hist
        gamma_curve = self.gamma_curve
        if self.ce_enable:
            if self.lo_max != 0 or self.hi_max != 0:
                gamma_curve = self.__compose_gamma_curve(self.__compute_stretch_curve(histogram), gamma_curve)
        if self.brightness != 0 or self.contrast != 1.0:
            gamma_curve = [
                (
                    x,
                    max(
                        0.0,
                        min(
                            65535.0,
                            (y - 32768) * self.contrast + 32768 + self.brightness,
                        ),
                    ),
                )
                for (x, y) in gamma_curve
            ]
        gm = bcm2835_isp_gamma()
        self.__fill_in_contrast_status(gm, gamma_curve)
        gamma = v4l2_ext_control()
        gamma.id = V4L2_CID.USER_BCM2835_ISP_GAMMA
        gamma.size = sizeof(bcm2835_isp_gamma)
        gamma.ptr = cast(pointer(gm), c_void_p)
        self.isp_in.set_ext_controls([gamma])

    def __adjust_setting_from_isp(self) -> None:
        buffer = self.isp_out_metadata.dequeue_buffer_nonblocking(v4l2_memory=V4L2_MEMORY.MMAP)
        if buffer is None:
            return

        stats: bcm2835_isp_stats = cast(buffer.mapped_buf, POINTER(bcm2835_isp_stats)).contents
        self.__calc_lux(stats)
        if self.do_agc:
            if self.agc_interval_count < AGC_INTERVAL:
                self.agc_interval_count += 1
            else:
                self.agc_interval_count = 0
                self.__agc(stats)

        if self.do_awb:
            self.__awb(stats)

        if self.do_contrast:
            self.__contrast_control(stats)
        if self.auto_focuser is not None:
            self.auto_focuser.process_contrast_metadata(stats)

        self.isp_out_metadata.queue_buffer(buffer.buf.index)

    def run(self) -> None:
        self.unicam.queue_all_buffers()
        if self.unicam_meta is not None:
            self.unicam_meta.queue_all_buffers()
        self.isp_out_high.queue_all_buffers()
        self.isp_out_metadata.queue_all_buffers()

        self.unicam.start_streaming()
        if self.unicam_meta is not None:
            self.unicam_meta.start_streaming()
        self.isp_in.start_streaming()
        self.isp_out_high.start_streaming()
        self.isp_out_metadata.start_streaming()

        while self._is_running():
            timeout = 1
            rlist, wlist, _ = select.select(
                [
                    self.unicam.device_fd,
                    self.isp_out_high.device_fd,
                    self.isp_out_metadata.device_fd,
                ]
                + ([] if self.unicam_meta is None else [self.unicam_meta.device_fd]),
                [self.isp_in.device_fd],
                [],
                timeout,
            )

            if len(rlist) == 0 and len(wlist) == 0:
                raise RuntimeError("Capture timeout")

            for r in rlist:
                if r == self.unicam.device_fd:
                    self.__unicam2isp()
                elif self.unicam_meta is not None and r == self.unicam_meta.device_fd:
                    self.__receive_unicam_meta()
                elif r == self.isp_out_high.device_fd:
                    self.__produce_image_from_isp()
                elif r == self.isp_out_metadata.device_fd:
                    self.__adjust_setting_from_isp()

            for w in wlist:
                if w == self.isp_in.device_fd:
                    self.__isp2unicam()

    def _new_pad(self) -> _PadBase[Frame[bytes]]:
        return _PadDiscardingOld()



import json
import os
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

from actfw_core.v4l2.video import Video, VideoPort  # type: ignore


class EnvironmentVariableNotSet(Exception):
    def __init__(self, name: str) -> None:
        self.name = name

    def __str__(self) -> str:
        return f"Environment variable {self.name} is not set. Perhaps the host agent is too old?"


@dataclass
class DeviceNode:
    path: Path
    label: Optional[str] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "DeviceNode":
        path = json.get("path", None)
        if path is None:
            raise ValueError("must have path property")

        if type(path) is not str:
            raise ValueError("path must be string")

        path = Path(path)

        label = json.get("label", None)
        if label is None:
            return DeviceNode(path=path, label=None)
        else:
            if type(label) is not str:
                raise ValueError("label must be string")
            return DeviceNode(path=path, label=label)


@dataclass
class DeviceInfo:
    type: str
    nodes: List[DeviceNode]

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "DeviceInfo":
        if "type" not in json.keys():
            raise ValueError("must have type property")
        if "nodes" not in json.keys():
            raise ValueError("must have nodes property")

        device_type = json["type"]
        if type(device_type) is not str:
            raise ValueError("type must be string")

        nodes = [DeviceNode.from_json(node) for node in json["nodes"]]
        return DeviceInfo(type=device_type, nodes=nodes)


@dataclass
class DeviceSupply:
    devices: List[DeviceInfo]

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "DeviceSupply":
        if "devices" not in json.keys():
            raise ValueError("must have devices property")

        devices = [DeviceInfo.from_json(d) for d in json["devices"]]
        return DeviceSupply(devices=devices)


def _get_env_str(name: str) -> str:
    ret = os.environ.get(name)
    if ret is None:
        raise EnvironmentVariableNotSet(name)
    return ret


def _get_env_int(name: str) -> int:
    ret = os.environ.get(name)
    if ret is None:
        raise EnvironmentVariableNotSet(name)
    return int(ret)


def get_actcast_protocol_version() -> str:
    "Protocol version of agent-app communication. Since ACTCAST_PROTOCOL_VERSION 1.0.0."
    return _get_env_str("ACTCAST_PROTOCOL_VERSION")


def get_actcast_device_id() -> str:
    "Device ID of the device running the actcast application. Since ACTCAST_PROTOCOL_VERSION 1.0.0."
    return _get_env_str("ACTCAST_DEVICE_ID")


def get_actcast_group_id() -> int:
    "Group ID that the device is belonging to. Since ACTCAST_PROTOCOL_VERSION 1.2.0."
    return _get_env_int("ACTCAST_GROUP_ID")


def get_actcast_act_id() -> int:
    "Act ID of the actcast application. Since ACTCAST_PROTOCOL_VERSION 1.0.0."
    return _get_env_int("ACTCAST_ACT_ID")


def get_actcast_instance_id() -> str:
    "ID that identifies the launch of the actcast application, like PID. Since ACTCAST_PROTOCOL_VERSION 1.0.0."
    return _get_env_str("ACTCAST_INSTANCE_ID")


def get_actcast_device_type() -> str:
    "Device type. Since ACTCAST_PROTOCOL_VERSION 1.1.0."
    return _get_env_str("ACTCAST_DEVICE_TYPE")


def get_act_settings_path() -> str:
    """
    Path of act settings.
    This is rather low level API.
    Use get_settings in Application to get act settings.
    Since ACTCAST_PROTOCOL_VERSION 1.0.0.
    """
    return _get_env_str("ACT_SETTINGS_PATH")


def get_actcast_command_sock() -> str:
    """
    Path of socket file to receive command from actcast agent.
    This is rather low level API.
    Use CommandServer to receive commands from actcast agent.
    Since ACTCAST_PROTOCOL_VERSION 1.0.0.
    """
    return _get_env_str("ACTCAST_COMMAND_SOCK")


def get_actcast_service_sock() -> str:
    """
    Path of socket file to send command to actcast agent.
    This is rather low level API.
    Use ServiceClient to send commands to actcast agent.
    Since ACTCAST_PROTOCOL_VERSION 1.0.0.
    """
    return _get_env_str("ACTCAST_SERVICE_SOCK")


def get_actcast_socks_server() -> Optional[str]:
    "URL of SOCKS5 proxy server. Since ACTCAST_PROTOCOL_VERSION 1.0.0."
    return os.environ.get("ACTCAST_SOCKS_SERVER")


def get_actcast_agent_simulator() -> Optional[str]:
    'Fixed value "actsim" if on actsim environment. Not set otherwise. Since ACTCAST_PROTOCOL_VERSION 1.1.0.'
    return os.environ.get("ACTCAST_AGENT_SIMULATOR")


def get_actcast_firmware_type() -> str:
    "Firmware type of the host. Since ACTCAST_PROTOCOL_VERSION 1.3.0."
    return _get_env_str("ACTCAST_FIRMWARE_TYPE")


def get_device_supply_path() -> str:
    """
    Path of device supply file.
    This is rather low level API.
    Use get_device_supply to get DeviceSupply object.
    Since ACTCAST_PROTOCOL_VERSION 1.3.0.
    """
    return _get_env_str("DEVICE_SUPPLY_PATH")


def get_device_supply() -> DeviceSupply:
    """
    Device Supply from actcast agent.
    Since ACTCAST_PROTOCOL_VERSION 1.3.0.
    """
    path = get_device_supply_path()
    with open(path) as f:
        return DeviceSupply.from_json(json.load(f))


def _get_camera_device_info(devs: DeviceSupply, default_image_source: Optional[str] = None) -> DeviceInfo:
    cameras = [dev for dev in devs.devices if dev.type == "camera"]
    if len(cameras) >= 1:
        if len(cameras) > 1:
            warnings.warn(
                "There are multiple camera candidates. Check your manifesto files.",
                RuntimeWarning,
            )
        info = cameras[0]
        sources = [node for node in info.nodes if node.label == "image_source"]
        if len(sources) >= 1:
            return info
        elif default_image_source is not None:
            if Path(default_image_source) in [node.path for node in info.nodes]:
                for node in info.nodes:
                    if node.path == Path(default_image_source):
                        node.label = "image_source"
                return info
            else:
                raise RuntimeError(f"default_image_source={default_image_source} is not mounted. Fix your manifesto files.")
        else:
            raise RuntimeError(
                "Not found image_source. Fix your manifesto files or give default_image_source to get_camera_device_info."
            )

    else:
        raise RuntimeError("Not found camera device. Fix your manifesto files.")


def get_camera_device_info(default_image_source: Optional[str] = None) -> DeviceInfo:
    """
    DeviceInfo for camera.
    Set `default_image_source` only if you write camera device path in manifesto files.
    Since ACTCAST_PROTOCOL_VERSION 1.3.0.
    """
    devs = get_device_supply()
    return _get_camera_device_info(devs, default_image_source)


def _list_video_devices() -> List[str]:
    devs = get_device_supply()
    cameras = [dev for dev in devs.devices if dev.type == "camera"]
    paths: List[str] = []
    for c in cameras:
        for node in filter(lambda n: str(n.path).startswith("/dev/video"), c.nodes):
            paths.append(str(node.path))
    return paths


def _find_specific_video_device(video_port: VideoPort) -> Optional[str]:
    devs = _list_video_devices()
    for dev in devs:
        try:
            # No `video.close()` leads to Capture timeout error.
            with Video(dev) as video:
                if video.query_capability() == video_port:
                    return dev
        except RuntimeError:
            pass
    return None


def find_usb_camera_device() -> Optional[str]:
    """
    Path of USB camera device.
    Since ACTCAST_PROTOCOL_VERSION 1.3.0.
    """
    return _find_specific_video_device(VideoPort.USB)


def find_csi_camera_device() -> Optional[str]:
    """
    Path of CSI camera device.
    Since ACTCAST_PROTOCOL_VERSION 1.3.0.
    """
    return _find_specific_video_device(VideoPort.CSI)



# flake8: noqa

import enum
from ctypes import *


class capability(Structure):
    _fields_ = [
        ("driver", c_uint8 * 16),
        ("card", c_uint8 * 32),
        ("bus_info", c_uint8 * 32),
        ("version", c_uint32),
        ("capabilities", c_uint32),
        ("device_caps", c_uint32),
        ("reserved", c_uint32 * 3),
    ]


class fmtdesc(Structure):
    _fields_ = [
        ("index", c_uint32),
        ("type", c_uint32),
        ("flags", c_uint32),
        ("description", c_uint8 * 32),
        ("pixelformat", c_uint32),
        ("reserved", c_uint32 * 4),
    ]


class frmsize_discrete(Structure):
    _fields_ = [
        ("width", c_uint32),
        ("height", c_uint32),
    ]


class frmsize_stepwise(Structure):
    _fields_ = [
        ("min_width", c_uint32),
        ("max_width", c_uint32),
        ("step_width", c_uint32),
        ("min_height", c_uint32),
        ("max_height", c_uint32),
        ("step_height", c_uint32),
    ]


class _frmsize_for_frmsizeenum(Union):
    _fields_ = [
        ("discrete", frmsize_discrete),
        ("stepwise", frmsize_stepwise),
    ]


class frmsizeenum(Structure):
    _anonymous_ = ("_frmsize",)
    _fields_ = [
        ("index", c_uint32),
        ("pixel_format", c_uint32),
        ("type", c_uint32),
        ("_frmsize", _frmsize_for_frmsizeenum),
        ("reserved", c_uint32 * 2),
    ]


class fract(Structure):
    _fields_ = [
        ("numerator", c_uint32),
        ("denominator", c_uint32),
    ]


class frmival_stepwise(Structure):
    _fields_ = [
        ("min", fract),
        ("max", fract),
        ("step", fract),
    ]


class _frmival_for_frmivalenum(Union):
    _fields_ = [
        ("discrete", fract),
        ("stepwise", frmival_stepwise),
    ]


class frmivalenum(Structure):
    _anonymous_ = ("_frmival",)
    _fields_ = [
        ("index", c_uint32),
        ("pixel_format", c_uint32),
        ("width", c_uint32),
        ("height", c_uint32),
        ("type", c_uint32),
        ("_frmival", _frmival_for_frmivalenum),
        ("reserved", c_uint32 * 2),
    ]


class pix_format(Structure):
    _fields_ = [
        ("width", c_uint32),
        ("height", c_uint32),
        ("pixelformat", c_uint32),
        ("field", c_uint32),
        ("bytesperline", c_uint32),
        ("sizeimage", c_uint32),
        ("colorspace", c_uint32),
        ("priv", c_uint32),
        ("flags", c_uint32),
        ("ycbcr_enc", c_uint32),
        ("quantization", c_uint32),
        ("xfer_func", c_uint32),
    ]


class plane_pix_format(Structure):
    _pack_ = 1
    _fields_ = [
        ("sizeimage", c_uint32),
        ("bytesperline", c_uint32),
        ("reserved", c_uint16 * 6),
    ]


class pix_format_mplane(Structure):
    _pack_ = 1
    _fields_ = [
        ("width", c_uint32),
        ("height", c_uint32),
        ("pixelformat", c_uint32),
        ("field", c_uint32),
        ("colorspace", c_uint32),
        ("plane_fmt", plane_pix_format * 8),  # VIDEO_MAX_PLANES
        ("num_planes", c_uint8),
        ("flags", c_uint8),
        ("ycbcr_enc", c_uint8),
        ("quantization", c_uint8),
        ("xfer_func", c_uint8),
        ("reserved", c_uint8 * 7),
    ]


class rect(Structure):
    _fields_ = [
        ("left", c_int32),
        ("top", c_int32),
        ("width", c_uint32),
        ("height", c_uint32),
    ]


class clip(Structure):
    pass


clip._fields_ = [
    ("c", rect),
    ("next", POINTER(clip)),
]


class window(Structure):
    _fields_ = [
        ("w", rect),
        ("field", c_uint32),
        ("chromakey", c_uint32),
        ("clips", POINTER(clip)),
        ("clipcount", c_uint32),
        ("bitmap", c_void_p),
        ("global_alpha", c_uint8),
    ]


class vbi_format(Structure):
    _fields_ = [
        ("sampling_rate", c_uint32),
        ("offset", c_uint32),
        ("samples_per_line", c_uint32),
        ("sample_format", c_uint32),
        ("start", c_int32 * 2),
        ("count", c_uint32 * 2),
        ("flags", c_uint32),
        ("reserved", c_uint32 * 2),
    ]


class sliced_vbi_format(Structure):
    _fields_ = [
        ("service_lines", 2 * (24 * c_ushort)),
        ("io_size", c_uint),
        ("reserved", c_uint * 2),
    ]


class sdr_format(Structure):
    _pack_ = 1
    _fields_ = [
        ("pixelformat", c_uint32),
        ("buffersize", c_uint32),
        ("reserved", c_uint8 * 24),
    ]


class meta_format(Structure):
    _pack_ = 1
    _fields_ = [
        ("dataformat", c_uint32),
        ("buffersize", c_uint32),
    ]


class _fmt_for_format(Union):
    _fields_ = [
        ("pix", pix_format),
        ("pix_mp", pix_format_mplane),
        ("win", window),
        ("vbi", vbi_format),
        ("sliced", sliced_vbi_format),
        ("sdr", sdr_format),
        ("meta", meta_format),
        ("raw_data", c_uint8 * 200),
    ]


class format(Structure):
    _fields_ = [
        ("type", c_uint32),
        ("fmt", _fmt_for_format),
    ]


class captureparm(Structure):
    _fields_ = [
        ("capability", c_uint32),
        ("capturemode", c_uint32),
        ("timeperframe", fract),
        ("extendedmode", c_uint32),
        ("readbufferbs", c_uint32),
        ("reserved", c_uint32 * 4),
    ]


class outputparm(Structure):
    _fields_ = [
        ("capability", c_uint32),
        ("outputmode", c_uint32),
        ("timeperframe", fract),
        ("extendedmode", c_uint32),
        ("writebuffers", c_uint32),
        ("reserved", c_uint32 * 4),
    ]


class _parm_for_streamparm(Union):
    _fields_ = [
        ("capture", captureparm),
        ("output", outputparm),
        ("raw_data", c_uint8 * 200),
    ]


class streamparm(Structure):
    _fields_ = [
        ("type", c_uint32),
        ("parm", _parm_for_streamparm),
    ]


class control(Structure):
    _fields_ = [
        ("id", c_uint32),
        ("value", c_int32),
    ]


class queryctrl(Structure):
    _fields_ = [
        ("id", c_uint32),
        ("type", c_uint32),
        ("name", c_uint8 * 32),
        ("minimum", c_int32),
        ("maximum", c_int32),
        ("step", c_int32),
        ("default_value", c_int32),
        ("flags", c_uint32),
        ("reserved", c_uint32 * 2),
    ]


class v4l2_query_ext_ctrl(Structure):
    _fields_ = [
        ("id", c_uint32),
        ("type", c_uint32),
        ("name", c_char * 32),
        ("minimum", c_longlong),
        ("maximum", c_longlong),
        ("step", c_int32),
        ("default_value", c_longlong),
        ("flags", c_uint32),
        ("elem_size", c_uint32),
        ("elems", c_uint32),
        ("nr_of_dims", c_uint32),
        ("dims", c_uint32 * 4),
        ("reserved", c_uint32 * 32),
    ]


class requestbuffers(Structure):
    _fields_ = [
        ("count", c_uint32),
        ("type", c_uint32),
        ("memory", c_uint32),
        ("reserved", c_uint32 * 2),
    ]


class exportbuffer(Structure):
    _fields_ = [
        ("type", c_uint32),
        ("index", c_uint32),
        ("plane", c_uint32),
        ("flags", c_uint32),
        ("fd", c_int32),
        ("reserved", c_uint32 * 11),
    ]


class timeval(Structure):
    # TODO: check
    _fields_ = [
        ("sec", c_long),
        ("usec", c_long),
    ]


class timecode(Structure):
    _fields_ = [
        ("type", c_uint32),
        ("flags", c_uint32),
        ("frames", c_uint8),
        ("seconds", c_uint8),
        ("minutes", c_uint8),
        ("hours", c_uint8),
        ("userbits", c_uint8 * 4),
    ]


class _m_for_plane(Union):
    _fields_ = [
        ("mem_offset", c_uint32),
        ("userptr", c_ulong),
        ("fd", c_int32),
    ]


class plane(Structure):
    _fields_ = [
        ("bytesused", c_uint32),
        ("length", c_uint32),
        ("m", _m_for_plane),
        ("data_offset", c_uint32),
        ("reserved", c_uint32 * 11),
    ]


class _m_for_buffer(Union):
    _fields_ = [
        ("offset", c_uint32),
        ("userptr", c_ulong),
        ("planes", POINTER(plane)),
        ("fd", c_int32),
    ]


class _fd_or_reserved(Union):
    _fields_ = [
        ("request_fd", c_int32),
        ("reserved", c_uint32),
    ]


class buffer(Structure):
    _fields_ = [
        ("index", c_uint32),
        ("type", c_uint32),
        ("bytesused", c_uint32),
        ("flags", c_uint32),
        ("field", c_uint32),
        ("timestamp", timeval),
        ("timecode", timecode),
        ("sequence", c_uint32),
        ("memory", c_uint32),
        ("m", _m_for_buffer),
        ("length", c_uint32),
        ("reserved2", c_uint32),
        ("reserved", _fd_or_reserved),
    ]


class mbus_framefmt(Structure):
    _fields_ = [
        ("width", c_uint32),
        ("height", c_uint32),
        ("code", c_uint32),
        ("field", c_uint32),
        ("colorspace", c_uint32),
        ("ycbcr_enc", c_uint16),
        ("quantization", c_uint16),
        ("xfer_func", c_uint16),
        ("reserved", c_uint16 * 11),
    ]


class subdev_format(Structure):
    _fields_ = [
        ("which", c_uint32),
        ("pad", c_uint32),
        ("format", mbus_framefmt),
        ("reserved", c_uint32 * 8),
    ]


class selection(Structure):
    _fields_ = [
        ("type", c_uint32),
        ("target", c_uint32),
        ("flags", c_uint32),
        ("r", rect),
        ("reserved", c_uint32 * 9),
    ]


class _value_for_ext_control(Union):
    _fields_ = [
        ("value", c_int32),
        ("value64", c_int64),
        ("string", c_char_p),
        ("p_u8", POINTER(c_uint8)),
        ("p_u16", POINTER(c_uint16)),
        ("p_u32", POINTER(c_uint32)),
        ("ptr", c_void_p),
    ]


class v4l2_ext_control(Structure):
    _pack_ = 1
    _anonymous_ = ("_value",)
    _fields_ = [
        ("id", c_uint32),
        ("size", c_uint32),
        ("reserved2", c_uint32 * 1),
        ("_value", _value_for_ext_control),
    ]


class _class_or_which_for_ext_controls(Union):
    _fields_ = [
        ("ctrl_class", c_uint32),
        ("which", c_uint32),
    ]


class v4l2_ext_controls(Structure):
    _anonymous_ = ("_class_or_which",)
    _fields_ = [
        ("_class_or_which", _class_or_which_for_ext_controls),
        ("count", c_uint32),
        ("error_idx", c_uint32),
        ("request_fd", c_int32),
        ("reserved", c_uint32),
        ("controls", POINTER(v4l2_ext_control)),
    ]


# ISP statistics structures.
# TODO: decide where to put the following definitions
DEFAULT_AWB_REGIONS_X = 16
DEFAULT_AWB_REGIONS_Y = 12

NUM_HISTOGRAMS = 2
NUM_HISTOGRAM_BINS = 128
AWB_REGIONS = DEFAULT_AWB_REGIONS_X * DEFAULT_AWB_REGIONS_Y
FLOATING_REGIONS = 16
AGC_REGIONS = 16
FOCUS_REGIONS = 12
CONTRAST_NUM_POINTS: int = 33


class bcm2835_isp_gain_format(enum.IntEnum):
    GAIN_FORMAT_U0P8_1 = 0
    GAIN_FORMAT_U1P7_0 = 1
    GAIN_FORMAT_U1P7_1 = 2
    GAIN_FORMAT_U2P6_0 = 3
    GAIN_FORMAT_U2P6_1 = 4
    GAIN_FORMAT_U3P5_0 = 5
    GAIN_FORMAT_U3P5_1 = 6
    GAIN_FORMAT_U4P10 = 7


class bcm2835_isp_black_level(Structure):
    _fields_ = [
        ("enabled", c_uint32),
        ("black_level_r", c_uint16),
        ("black_level_g", c_uint16),
        ("black_level_b", c_uint16),
        ("padding", c_uint8 * 2),
    ]


class bcm2835_isp_gamma(Structure):
    _fields_ = [
        ("enabled", c_uint32),
        ("x", c_uint16 * CONTRAST_NUM_POINTS),
        ("y", c_uint16 * CONTRAST_NUM_POINTS),
    ]


class bcm2835_isp_stats_hist(Structure):
    _fields_ = [
        ("r_hist", c_uint32 * NUM_HISTOGRAM_BINS),
        ("g_hist", c_uint32 * NUM_HISTOGRAM_BINS),
        ("b_hist", c_uint32 * NUM_HISTOGRAM_BINS),
    ]


class bcm2835_isp_stats_region(Structure):
    _fields_ = [
        ("counted", c_uint32),
        ("noncounted", c_uint32),
        ("r_sum", c_uint64),
        ("g_sum", c_uint64),
        ("b_sum", c_uint64),
    ]


class bcm2835_isp_stats_focus(Structure):
    _fields_ = [
        ("contrast_val", c_uint64 * 2 * 2),
        ("contrast_val_num", c_uint32 * 2 * 2),
    ]


class bcm2835_isp_stats_contrast(Structure):
    _fields_ = [
        ("points", c_uint16 * 2 * CONTRAST_NUM_POINTS),
        ("brightness", c_double),
        ("contrast", c_double),
    ]


class bcm2835_isp_stats(Structure):
    _fields_ = [
        ("version", c_uint32),
        ("size", c_uint32),
        ("hist", bcm2835_isp_stats_hist * NUM_HISTOGRAMS),
        ("awb_stats", bcm2835_isp_stats_region * AWB_REGIONS),
        ("floating_stats", bcm2835_isp_stats_region * FLOATING_REGIONS),
        ("agc_stats", bcm2835_isp_stats_region * AGC_REGIONS),
        ("focus_stats", bcm2835_isp_stats_focus * FOCUS_REGIONS),
    ]


class bcm2835_isp_lens_shading(Structure):
    _fields_ = [
        ("enabled", c_uint32),
        ("grid_cell_size", c_uint32),
        ("grid_width", c_uint32),
        ("grid_stride", c_uint32),
        ("grid_height", c_uint32),
        ("dmabuf", c_int32),
        ("ref_transform", c_uint32),
        ("corner_sampled", c_uint32),
        ("gain_format", c_uint32),
    ]



import base64
import copy
import os
import socket
from pathlib import Path
from typing import Optional

import OpenSSL.crypto

from ..compat.queue import SimpleQueue
from ..schema.agent_app_protocol import ServiceRequest, ServiceResponse, Status
from ..util.result import ResultTuple
from ..util.thread import LoopThread


class AgentAppProtocolServiceServer:
    _path: Path
    _pkey: OpenSSL.crypto.PKey
    _thread: LoopThread
    _listener: Optional[socket.socket]

    def __init__(self, path: Path, pkey: OpenSSL.crypto.PKey) -> None:
        self._path = path
        self._pkey = pkey
        dummy_ch: SimpleQueue = SimpleQueue()
        self._thread = LoopThread(dummy_ch, self._loop_body)
        self._listener = None

    def path(self) -> Path:
        return self._path

    def startup(self) -> None:
        try:
            os.unlink(self._path)
        except FileNotFoundError:
            pass
        self._listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._listener.bind(str(self._path))
        self._path.chmod(0o664)
        self._listener.settimeout(1)
        self._listener.listen(1)

        self._thread.startup()

    def teardown(self) -> None:
        self._thread.teardown()

    def join(self) -> None:
        self._thread.join()

    def _loop_body(self) -> None:
        assert self._listener is not None

        try:
            stream, _ = self._listener.accept()
        except socket.timeout:
            return

        request, err = ServiceRequest.parse(stream)

        if request is None:
            raise RuntimeError("couldn't parse a request from actcast agent: `ServiceRequest.parse()` failed")

        if err is not None:
            response = ServiceResponse(
                copy.copy(request.id_),
                Status.GENERAL_ERROR,
                b"",
            )
            stream.sendall(response.to_bytes())
            return

        # Currently, only RS_256 is supported.
        response = self._handle_rs_256(request)

        stream.sendall(response.to_bytes())

    def _handle_rs_256(self, request: ServiceRequest) -> ServiceResponse:
        bs, err = _base64_decode_url_safe_no_pad(request.data)
        if (err is not None) or (bs is None):
            return ServiceResponse(
                copy.copy(request.id_),
                Status.GENERAL_ERROR,
                b"",
            )

        bs = OpenSSL.crypto.sign(self._pkey, bs, "sha256")
        bs = _base64_encode_url_safe_no_pad(bs)
        return ServiceResponse(
            copy.copy(request.id_),
            Status.OK,
            bs,
        )


def _base64_decode_url_safe_no_pad(data: bytes) -> ResultTuple[bytes, bool]:
    """
    Decode with config `base64::URL_SAFE_NO_PAD`.
    c.f. https://docs.rs/base64/0.13.0/base64/constant.URL_SAFE_NO_PAD.html
    """
    missing_pad_len = -(len(data) % -4)
    data_ = data + b"=" * missing_pad_len
    try:
        return base64.urlsafe_b64decode(data_), None
    except Exception:
        return None, True


def _base64_encode_url_safe_no_pad(data: bytes) -> bytes:
    """
    Encode with config `base64::URL_SAFE_NO_PAD`.
    c.f. https://docs.rs/base64/0.13.0/base64/constant.URL_SAFE_NO_PAD.html
    """
    return base64.urlsafe_b64encode(data).rstrip(b"=")



from abc import ABC, abstractmethod
from queue import Empty, Queue
from typing import Generic, Optional, Tuple, TypeVar

T = TypeVar("T")


class _PadBase(ABC, Generic[T]):
    _queue: "Queue[T]"

    @abstractmethod
    def empty(self) -> bool:
        pass

    @abstractmethod
    def put(
        self,
        item: T,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> None:
        pass

    @abstractmethod
    def get(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> T:
        return self._queue.get(block=block, timeout=timeout)

    def into_pad_pair(self) -> "Tuple[_PadOut[T], _PadIn[T]]":
        return (_PadOut(self), _PadIn(self))


class _PadBlocking(_PadBase[T]):
    _queue: "Queue[T]"

    def __init__(self) -> None:
        self._queue = Queue(1)

    def empty(self) -> bool:
        return self._queue.empty()

    def put(
        self,
        item: T,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> None:
        self._queue.put(item, block=block, timeout=timeout)

    def get(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> T:
        return self._queue.get(block=block, timeout=timeout)


class _PadDiscardingOld(_PadBase[T]):
    _queue: "Queue[T]"

    def __init__(self) -> None:
        self._queue = Queue(1)

    def empty(self) -> bool:
        return self._queue.empty()

    def put(
        self,
        item: T,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> None:
        # First, make the queue empty.
        # Notice that only the owner of `_PadIn` can `put()`, but the owner can call `put()` cuncurrently.
        try:
            self.get(block=False)
        except Empty:
            pass
        # Then put.
        self._queue.put(item, block=block, timeout=timeout)

    def get(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> T:
        return self._queue.get(block=block, timeout=timeout)


class _PadIn(Generic[T]):
    _pad: _PadBase[T]

    def __init__(self, pad: _PadBase[T]) -> None:
        self._pad = pad

    def put(
        self,
        item: T,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> None:
        self._pad.put(item, block=block, timeout=timeout)


class _PadOut(Generic[T]):
    _pad: _PadBase[T]

    def __init__(self, pad: _PadBase[T]) -> None:
        self._pad = pad

    def empty(self) -> bool:
        return self._pad.empty()

    def get(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> T:
        return self._pad.get(block=block, timeout=timeout)



from threading import Thread
from typing import Callable

from ..compat.queue import SimpleQueue


class LoopThread:
    _error_ch: SimpleQueue
    _loop_body: Callable[[], None]
    _allow_extra_loop_for_teardown: bool
    _thread: Thread
    _teardown_ch: SimpleQueue

    def __init__(
        self,
        error_ch: SimpleQueue,
        loop_body: Callable[[], None],
        allow_extra_loop_for_teardown: bool = False,
        daemon: bool = False,
    ) -> None:
        self._error_ch = error_ch
        self._teardown_ch = SimpleQueue()
        self._loop_body = loop_body
        self._thread = Thread(target=self._loop, daemon=daemon)
        self._allow_extra_loop_for_teardown = allow_extra_loop_for_teardown

    def startup(self) -> None:
        self._thread.start()

    def teardown(self) -> None:
        self._teardown_ch.put(None)

    def join(self) -> None:
        self._thread.join()

    def _loop(self) -> None:
        try:
            while self._teardown_ch.empty():
                self._loop_body()
            if self._allow_extra_loop_for_teardown:
                self._loop_body()
        except Exception as e:
            self._error_ch.put(e)



# type: ignore
# flake8: noqa
import errno
import os
from ctypes import c_char_p, get_errno
from fcntl import ioctl

from actfw_core.linux.ioctl import _IOW, _IOWR
from actfw_core.linux.types import dma_heap_allocation_data

_DMA_HEAP_IOCTL_ALLOC = _IOWR("H", 0, dma_heap_allocation_data)
_DMA_BUF_SET_NAME = _IOW("b", 1, c_char_p)


class DMAHeap(object):
    def __init__(self, node="/dev/dma_heap/linux,cma"):
        self.heap_fd = os.open(node, os.O_RDWR)

    def alloc(self, name: str, size: int) -> int:
        alloc = dma_heap_allocation_data()
        alloc.len = size
        alloc.fd_flags = os.O_CLOEXEC | os.O_RDWR
        result = ioctl(self.heap_fd, _DMA_HEAP_IOCTL_ALLOC, alloc)
        if -1 == result:
            raise RuntimeError(f"ioctl(_DMA_HEAP_IOCTL_ALLOC){errno.errorcode[get_errno()]}")

        result = ioctl(alloc.fd, _DMA_BUF_SET_NAME, name)

        if -1 == result:
            raise RuntimeError(f"ioctl(_DMA_BUF_SET_NAME){errno.errorcode[get_errno()]}")

        return alloc.fd



import json
import os
import signal
import sys
import time
from types import FrameType
from typing import Any, Dict, Iterable, List, Optional

from actfw_core.task import Task


class SettingSchema:
    def __init__(
        self,
        title: str,
        description: str,
        type_: type,
        default: Any = None,
        ui_type: Optional[str] = None,
    ):
        self.title = title
        self.description = description
        self.type = type_
        self.default = default
        self.ui_type = ui_type

    @staticmethod
    def decoder(obj: Any) -> Any:
        if "title" in obj and "description" in obj and "type" in obj:
            return SettingSchema(
                obj["title"],
                obj["description"],
                SettingSchema.infertype(obj["type"]),
                obj.get("default", None),
                obj.get("x-ui-type"),
            )
        return obj

    @staticmethod
    def infertype(typestring: str) -> type:
        if typestring == "number":
            return float
        elif typestring == "integer":
            return int
        elif typestring == "boolean":
            return bool
        else:
            return str


class AppSettings:
    def __init__(self, settings: Dict[str, Any], schema: Dict[str, SettingSchema]):
        self.settings = settings
        self.schema = schema

    def __getattr__(self, name: str) -> Any:
        if name in self.settings:
            if isinstance(self.settings[name], self.schema[name].type):
                return self.settings[name]
            elif self.schema[name].default is not None:
                print(
                    f"Invalid type for setting:{name}. Using schema default value.",
                    file=sys.stderr,
                    flush=True,
                )
                return self.schema[name].default
            else:
                print(f"Invalid type of {name}: {type(name)}", file=sys.stderr, flush=True)
        elif name in self.schema:
            print(
                f"Setting:{name} not found. Using schema default value.",
                file=sys.stderr,
                flush=True,
            )
            return self.schema[name].default
        else:
            raise AttributeError(f"{name} is not found in settings.")

    def __repr__(self) -> str:
        values = []
        for key, schema in self.schema.items():
            if schema.ui_type == "password":
                values.append(f"{key}=***")
            else:
                values.append(f"{key}={getattr(self, key)}")
        return f"AppSettings({', '.join(values)})"


class Application:
    running: bool
    tasks: List[Task]
    settings: Optional[Dict[str, Any]]

    """Actcast Application"""

    def __init__(
        self,
        stop_by_signals: Iterable[signal.Signals] = (signal.SIGINT, signal.SIGTERM),
    ) -> None:
        self.running = True
        for sig in stop_by_signals:
            signal.signal(sig, self._handler)  # type: ignore[arg-type]
        self.tasks = []
        self.settings = None
        env = "ACT_SETTINGS_PATH"
        if env in os.environ:
            try:
                with open(os.environ[env]) as f:
                    self.settings = json.load(f)
            except FileNotFoundError:
                pass

    def _handler(self, sig: signal.Signals, frame: FrameType) -> None:
        self.stop()

    def get_settings(self, default_settings: Dict[str, Any]) -> Dict[str, Any]:
        """

        Get given Act settings.

        Args:
            default_settings (dict): default settings

        Returns:
            dict: updated settings

        Notes:
            Copy default_settings and overwrite it by given Act settings.

        """
        if not isinstance(default_settings, dict):
            raise TypeError("default_settings must be dict.")
        settings = default_settings.copy()
        if self.settings is not None:
            settings.update(self.settings)
        return settings

    def get_app_settings(self, path: str = "setting_schema.json") -> AppSettings:
        """
        Get application settings from schema.
        Args:
            path (str): path to schema file
        Returns:
            AppSettings: application settings
        """
        with open(path) as f:
            self.schema = json.load(f, object_hook=SettingSchema.decoder)["properties"]
        return AppSettings(self.get_settings({}), self.schema)

    def register_task(self, task: Task) -> None:
        """

        Register the application task.

        Args:
            task (:class:`~actfw_core.task.Task`): task
        """
        if not issubclass(type(task), Task):
            raise TypeError("type(task) must be a subclass of actfw_core.task.Task.")
        self.tasks.append(task)

    def run(self) -> None:
        """Start application"""
        for task in self.tasks:
            task.start()

        try:
            while self.running:
                time.sleep(1)
        except KeyboardInterrupt:
            pass

        for task in self.tasks:
            task.stop()
        for task in self.tasks:
            task.join()

    def stop(self) -> None:
        """Stop application"""
        self.running = False



from typing import Generic, TypeVar

from .consumer import _ConsumerMixin
from .producer import _ProducerMixin
from .task import Task

T_OUT = TypeVar("T_OUT")
T_IN = TypeVar("T_IN")


class Pipe(Generic[T_OUT, T_IN], Task, _ProducerMixin[T_OUT], _ConsumerMixin[T_IN]):
    """Straightforward Pipeline Task."""

    def __init__(self) -> None:
        """"""
        Task.__init__(self)
        _ProducerMixin.__init__(self)
        _ConsumerMixin.__init__(self)

    def cleanup(self) -> None:
        """
        Perform cleanup before exiting.

        This method is executed at the end of `run`, even if an exception is raised.
        Since a long-running cleanup may cause the entire process to be terminated with SIGKILL,
        it must be implemented to complete quickly.
        """
        pass

    def run(self) -> None:
        """Run and start the activity"""
        try:
            for i in self._inlet():
                o = self.proc(i)
                self._outlet(o)
                if not self._is_running():
                    break
        finally:
            self.cleanup()

    def proc(self, i: T_IN) -> T_OUT:
        """
        Pipeline Task Processor
        """
        raise NotImplementedError("'proc' must be overridden.")



from queue import Full
from typing import Generic, TypeVar

from .consumer import _ConsumerMixin
from .producer import _ProducerMixin
from .task import Task

T = TypeVar("T")


class Tee(Generic[T], Task, _ProducerMixin[T], _ConsumerMixin[T]):
    """Tee Task."""

    def __init__(self) -> None:
        """"""
        Task.__init__(self)
        _ProducerMixin.__init__(self)
        _ConsumerMixin.__init__(self)

    def _outlet(self, o: T) -> bool:
        while self._is_running():
            for out_queue in self.out_queues:
                try:
                    out_queue.put(o, timeout=1)
                except Full:
                    pass
            return True
        return False

    def run(self) -> None:
        """Run and start the activity"""
        for i in self._inlet():
            o = i
            self._outlet(o)
            if not self._is_running():
                break



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

from actfw_core.v4l2.types import bcm2835_isp_stats
from actfw_core.v4l2.video import VideoBuffer  # type: ignore


def clamp(x: float, low: float, high: float) -> float:
    return max(low, min(x, high))


PDAF_STATS_ROWS = 12
PDAF_STATS_COLS = 16
CONTRAST_STATS_ROWS = 3
CONTRAST_STATS_COLS = 4
MAX_WINDOWS = 10


class AfRange(Enum):
    AfRangeNormal = "normal"
    AfRangeMacro = "macro"
    AfRangeFull = "full"
    AfRangeMax = "max"


class AfSpeed(Enum):
    AfSpeedNormal = "normal"
    AfSpeedFast = "fast"
    AfSpeedMax = "max"


class AfMode(Enum):
    AfModeManual = 0
    AfModeAuto = 1
    AfModeContinuous = 2


class AfPause(Enum):
    AfPauseImmediate = 0
    AfPauseDeferred = 1
    AfPauseResume = 2


class AfState(Enum):
    Idle = 0
    Scanning = 1
    Focused = 2
    Failed = 3


class AfPauseState(Enum):
    Running = 0
    Pausing = 1
    Paused = 2


class AfStatus:
    def __init__(self) -> None:
        self.state: AfState = AfState.Idle
        self.pauseState: AfPauseState = AfPauseState.Running
        self.lensSetting: Optional[int] = None


class ScanState(Enum):
    Idle = 0
    Trigger = 1
    Pdaf = 2
    Coarse = 3
    Fine = 4
    Settle = 5


"""
Default values for parameters. All may be overridden in the tuning file.
any of these values are sensor- or module-dependent; the defaults here
assume IMX708 in a Raspberry Pi V3 camera with the standard lens.
"""


class RangeDependentParams:
    def __init__(self) -> None:
        self.focusMin = 0.0
        self.focusMax = 12.0
        self.focusDefault = 1.0

    def read(self, params: Dict[str, Any]) -> None:
        self.focusMin = params.get("min", self.focusMin)
        self.focusMax = params.get("max", self.focusMax)
        self.focusDefault = params.get("default", self.focusDefault)


class SpeedDependentParams:
    def __init__(self) -> None:
        self.stepCoarse = 1.0
        self.stepFine = 0.25
        self.contrastRatio = 0.75
        self.pdafGain = -0.02
        self.pdafSquelch = 0.125
        self.maxSlew = 2.0
        self.pdafFrames = 20
        self.dropoutFrames = 6
        self.stepFrames = 4

    def read(self, params: Dict[str, Any]) -> None:
        self.stepCoarse = params.get("step_coarse", self.stepCoarse)
        self.stepFine = params.get("step_fine", self.stepFine)
        self.contrastRatio = params.get("contrast_ratio", self.contrastRatio)
        self.pdafGain = params.get("pdaf_gain", self.pdafGain)
        self.pdafSquelch = params.get("pdaf_squelch", self.pdafSquelch)
        self.maxSlew = params.get("max_slew", self.maxSlew)
        self.pdafFrames = params.get("pdaf_frames", self.pdafFrames)
        self.dropoutFrames = params.get("dropout_frames", self.dropoutFrames)
        self.stepFrames = params.get("step_frames", self.stepFrames)


class Pwl:
    def __init__(self, points: List[float]) -> None:
        assert len(points) % 2 == 0
        self.points_ = [[points[i], points[i + 1]] for i in range(0, len(points), 2)]

    def findSpan(self, x: float, span: int) -> int:
        """
        Pwls are generally small, so linear search may well be faster than
        binary, though could review this if large Pwls start turning up.
        """
        lastSpan = len(self.points_) - 2

        # Ensure span is within valid range
        span = max(0, min(lastSpan, span))

        # Adjust span based on x
        # assuming point[0] is x value
        while span < lastSpan and x >= self.points_[span + 1][0]:
            span += 1
        while span != 0 and x < self.points_[span][0]:
            span -= 1

        return span

    def eval(self, x: float) -> float:
        # Evaluate the piecewise linear function
        index = self.findSpan(x, len(self.points_) // 2 - 1)

        return self.points_[index][1] + (x - self.points_[index][0]) * (self.points_[index + 1][1] - self.points_[index][1]) / (
            self.points_[index + 1][0] - self.points_[index][0]
        )

    def inverse_eval(self, y: int) -> Optional[float]:
        # Find the corresponding x for a given y
        for i in range(len(self.points_) - 1):
            y1, y2 = self.points_[i][1], self.points_[i + 1][1]
            if (y1 <= y <= y2) or (y2 <= y <= y1):
                # Linear interpolation to find x
                x1, x2 = self.points_[i][0], self.points_[i + 1][0]
                return x1 + (y - y1) * (x2 - x1) / (y2 - y1)
        return None


class CfgParams:
    def __init__(self, params: Dict[str, Any]) -> None:
        self.confEpsilon = 8
        self.confThresh = 16
        self.confClip = 512
        self.skipFrames = 5
        self.ranges = {
            AfRange.AfRangeNormal: RangeDependentParams(),
            AfRange.AfRangeMacro: RangeDependentParams(),
            AfRange.AfRangeFull: RangeDependentParams(),
        }
        self.speeds = {
            AfSpeed.AfSpeedNormal: SpeedDependentParams(),
            AfSpeed.AfSpeedFast: SpeedDependentParams(),
        }
        self.read(params)

    def read(self, params: Dict[str, Any]) -> None:
        if "ranges" in params:
            rr = params["ranges"]
            self.ranges[AfRange.AfRangeNormal].read(rr.get("normal", {}))
            self.ranges[AfRange.AfRangeMacro].read(rr.get("macro", {}))

            self.ranges[AfRange.AfRangeFull].focusMin = min(
                self.ranges[AfRange.AfRangeNormal].focusMin,
                self.ranges[AfRange.AfRangeMacro].focusMin,
            )
            self.ranges[AfRange.AfRangeFull].focusMax = max(
                self.ranges[AfRange.AfRangeNormal].focusMax,
                self.ranges[AfRange.AfRangeMacro].focusMax,
            )
            self.ranges[AfRange.AfRangeFull].focusDefault = self.ranges[AfRange.AfRangeNormal].focusDefault
            self.ranges[AfRange.AfRangeFull].read(rr.get("full", {}))

        if "speeds" in params:
            ss = params["speeds"]
            self.speeds[AfSpeed.AfSpeedNormal].read(ss.get("normal", {}))
            self.speeds[AfSpeed.AfSpeedFast].read(ss.get("fast", {}))

        self.confEpsilon = params.get("conf_epsilon", self.confEpsilon)
        self.confThresh = params.get("conf_thresh", self.confThresh)
        self.confClip = params.get("conf_clip", self.confClip)
        self.skipFrames = params.get("skip_frames", self.skipFrames)

        if "map" in params:
            self.map = Pwl(params["map"])
        else:
            # Default mapping from dioptres to hardware setting
            DefaultMapX0 = 0.0
            DefaultMapY0 = 445.0
            DefaultMapX1 = 15.0
            DefaultMapY1 = 925.0
            self.map = Pwl([DefaultMapX0, DefaultMapY0, DefaultMapX1, DefaultMapY1])


class Rectangle:
    def __init__(self, xpos: float, ypos: float, w: float, h: float) -> None:
        self.x = xpos
        self.y = ypos
        self.width = w
        self.height = h


class PdafData:
    def __init__(self, conf: int, phase: int) -> None:
        self.conf = conf
        self.phase = phase


class PdafRegions:
    def __init__(self, height: int, width: int) -> None:
        self.height = height
        self.width = width
        self.pdaf_grid = [[PdafData(0, 0) for _ in range(width)] for _ in range(height)]
        self.numRegions = height * width


class ScanRecord:
    def __init__(self, focus: float, contrast: float, phase: float, conf: float) -> None:
        self.focus = focus
        self.contrast = contrast
        self.phase = phase
        self.conf = conf


class RegionWeights:
    def __init__(self, rows: int, cols: int) -> None:
        self.rows = rows
        self.cols = cols
        self.sum = 0.0
        self.w = [0.0] * rows * cols


class AutoFocuserBase(metaclass=ABCMeta):
    """
    Autofocus control algorithm
    https://github.com/raspberrypi/libcamera/blob/v0.3.2%2Brpt20240927/src/ipa/rpi/controller/rpi/af.cpp
    """

    def __init__(
        self,
        afrange: AfRange = AfRange.AfRangeNormal,
        afspeed: AfSpeed = AfSpeed.AfSpeedNormal,
        afmode: AfMode = AfMode.AfModeContinuous,
    ) -> None:
        self.bitdepth = 10  # hardcode
        self.afrange = afrange
        self.afspeed = afspeed
        self.afmode = afmode
        self.pause_flag = False
        self.statsRegion = Rectangle(0, 0, 0, 0)
        self.windows: List[Rectangle] = []
        self.phaseWeights = RegionWeights(rows=PDAF_STATS_ROWS, cols=PDAF_STATS_COLS)
        self.contrastWeights = RegionWeights(rows=CONTRAST_STATS_ROWS, cols=CONTRAST_STATS_COLS)
        # self.scan_state = ScanState.Idle
        self.scan_state = ScanState.Trigger if afmode == AfMode.AfModeContinuous else ScanState.Idle
        self.initted = False
        self.ftarget = -1.0
        self.fsmooth = -1.0
        self.prev_contrast = 0.0
        self.skip_count = 0
        self.step_count = 0
        self.drop_count = 0
        self.scan_max_contrast = 0.0
        self.scan_min_contrast = 1e9
        self.scan_max_index = 0
        self.scan_data: List[ScanRecord] = []
        self.report_state = AfState.Idle
        self.afstatus = AfStatus()

    def __update_lens_position(self) -> None:
        if self.scan_state.value >= ScanState.Pdaf.value:
            self.ftarget = clamp(
                self.ftarget,
                self.cfg.ranges[self.afrange].focusMin,
                self.cfg.ranges[self.afrange].focusMax,
            )
        if self.initted:
            self.fsmooth = clamp(
                self.ftarget,
                self.fsmooth - self.cfg.speeds[self.afspeed].maxSlew,
                self.fsmooth + self.cfg.speeds[self.afspeed].maxSlew,
            )
        else:
            self.fsmooth = self.ftarget
            self.initted = True
            self.skip_count = self.cfg.skipFrames

    def __startAF(self) -> None:
        if self.cfg.speeds[self.afspeed].dropoutFrames > 0 and (
            self.afmode == AfMode.AfModeContinuous or self.cfg.speeds[self.afspeed].pdafFrames > 0
        ):
            if not self.initted:
                self.ftarget = self.cfg.ranges[self.afrange].focusDefault
                self.__update_lens_position()
            self.step_count = 0 if self.afmode == AfMode.AfModeContinuous else self.cfg.speeds[self.afspeed].pdafFrames
            self.scan_state = ScanState.Pdaf
            self.scan_data.clear()
            self.drop_count = 0
            self.report_state = AfState.Scanning
        else:
            self.__start_programmed_scan()

    def __start_programmed_scan(self) -> None:
        self.ftarget = self.cfg.ranges[self.afrange].focusMin
        self.__update_lens_position()
        self.scan_state = ScanState.Coarse
        self.scan_max_contrast = 0.0
        self.scan_min_contrast = 1e9
        self.scan_max_index = 0
        self.scan_data.clear()
        self.step_count = self.cfg.speeds[self.afspeed].stepFrames
        self.report_state = AfState.Scanning

    def __doPDAF(self, phase: float, conf: float) -> None:
        phase *= self.cfg.speeds[self.afspeed].pdafGain

        if self.afmode == AfMode.AfModeContinuous:
            phase *= conf / (conf + self.cfg.confEpsilon)
            if abs(phase) < self.cfg.speeds[self.afspeed].pdafSquelch:
                a = phase / self.cfg.speeds[self.afspeed].pdafSquelch
                phase *= a * a
        else:
            if self.step_count >= self.cfg.speeds[self.afspeed].stepFrames:
                if abs(phase) < self.cfg.speeds[self.afspeed].pdafSquelch:
                    self.step_count = self.cfg.speeds[self.afspeed].stepFrames
            else:
                phase *= self.step_count / self.cfg.speeds[self.afspeed].stepFrames
        if phase < -1 * self.cfg.speeds[self.afspeed].maxSlew:
            phase = -1 * self.cfg.speeds[self.afspeed].maxSlew
            if self.ftarget <= self.cfg.ranges[self.afrange].focusMin:
                self.report_state = AfState.Failed
            else:
                self.report_state = AfState.Scanning
        elif phase > self.cfg.speeds[self.afspeed].maxSlew:
            phase = self.cfg.speeds[self.afspeed].maxSlew
            if self.ftarget >= self.cfg.ranges[self.afrange].focusMax:
                self.report_state = AfState.Failed
            else:
                self.report_state = AfState.Scanning
        else:
            self.report_state = AfState.Focused
        self.ftarget = self.fsmooth + phase

    def __early_termination_byphase(self, phase: float) -> bool:
        if len(self.scan_data) > 0 and self.scan_data[-1].conf >= self.cfg.confEpsilon:
            oldFocus = self.scan_data[-1].focus
            oldPhase = self.scan_data[-1].phase
            if (self.ftarget - oldFocus) * (phase - oldPhase) > 0.0:
                param = phase / (phase - oldPhase)
                if -3.0 <= param and param <= 3.5:
                    self.ftarget += param * (oldFocus - self.ftarget)
                    return True
        return False

    def __find_peak(self, i: int) -> float:
        f = self.scan_data[i].focus
        if i > 0 and i + 1 < len(self.scan_data):
            dropLo = self.scan_data[i].contrast - self.scan_data[i - 1].contrast
            dropHi = self.scan_data[i].contrast - self.scan_data[i + 1].contrast
            if 0.0 <= dropLo and dropLo < dropHi:
                param = 0.3125 * (1.0 - dropLo / dropHi) * (1.6 - dropLo / dropHi)
                f += param * (self.scan_data[i - 1].focus - f)
            elif 0.0 <= dropHi and dropHi < dropLo:
                param = 0.3125 * (1.0 - dropHi / dropLo) * (1.6 - dropHi / dropLo)
                f += param * (self.scan_data[i + 1].focus - f)
        return f

    def __doScan(self, contrast: float, phase: float, conf: float) -> None:
        if len(self.scan_data) == 0 or contrast > self.scan_max_contrast:
            self.scan_max_contrast = contrast
            self.scan_max_index = len(self.scan_data)
        if contrast < self.scan_min_contrast:
            self.scan_min_contrast = contrast
        self.scan_data.append(ScanRecord(self.ftarget, contrast, phase, conf))

        if self.scan_state == ScanState.Coarse:
            if (
                self.ftarget >= self.cfg.ranges[self.afrange].focusMax
                or contrast < self.cfg.speeds[self.afspeed].contrastRatio * self.scan_max_contrast
            ):
                self.ftarget = min(
                    self.ftarget,
                    self.__find_peak(self.scan_max_index) + 2.0 * self.cfg.speeds[self.afspeed].stepFine,
                )
                self.scan_state = ScanState.Fine
                self.scan_data.clear()
            else:
                self.ftarget += self.cfg.speeds[self.afspeed].stepCoarse
        else:
            if (
                self.ftarget <= self.cfg.ranges[self.afrange].focusMin
                or len(self.scan_data) >= 5
                or contrast < self.cfg.speeds[self.afspeed].contrastRatio * self.scan_max_contrast
            ):
                self.ftarget = self.__find_peak(self.scan_max_index)
                self.scan_state = ScanState.Settle
            else:
                self.ftarget -= self.cfg.speeds[self.afspeed].stepFine

        self.step_count = 0 if self.ftarget == self.fsmooth else self.cfg.speeds[self.afspeed].stepFrames

    def __doAF(self, contrast: float, phase: float, conf: float) -> None:
        if self.skip_count > 0:
            self.skip_count -= 1
            return
        if self.scan_state == ScanState.Pdaf:
            if conf > (1.0 if self.drop_count != 0 else 0.25) * self.cfg.confEpsilon:
                self.__doPDAF(phase, conf)
                if self.step_count > 0:
                    self.step_count -= 1
                elif self.afmode != AfMode.AfModeContinuous:
                    self.scan_state = ScanState.Idle
                self.drop_count = 0
            elif self.drop_count + 1 == self.cfg.speeds[self.afspeed].dropoutFrames:
                self.drop_count += 1
                self.__start_programmed_scan()
        elif self.scan_state.value >= ScanState.Coarse.value and self.fsmooth == self.ftarget:
            if self.step_count > 0:
                self.step_count -= 1
            elif self.scan_state == ScanState.Settle:
                if (
                    self.prev_contrast >= self.cfg.speeds[self.afspeed].contrastRatio * self.scan_max_contrast
                    and self.scan_min_contrast <= self.cfg.speeds[self.afspeed].contrastRatio * self.scan_max_contrast
                ):
                    self.report_state = AfState.Focused
                else:
                    self.report_state = AfState.Failed
                if (
                    self.afmode == AfMode.AfModeContinuous
                    and (not self.pause_flag)
                    and self.cfg.speeds[self.afspeed].dropoutFrames > 0
                ):
                    self.scan_state = ScanState.Pdaf
                else:
                    self.scan_state = ScanState.Idle
                self.scan_data.clear()
            elif conf >= self.cfg.confEpsilon and self.__early_termination_byphase(phase):
                self.scan_state = ScanState.Settle
                self.step_count = 0 if self.afmode == AfMode.AfModeContinuous else self.cfg.speeds[self.afspeed].stepFrames
            else:
                self.__doScan(contrast, phase, conf)

    def __compute_weights(self, weights: RegionWeights, rows: int, cols: int) -> None:
        weights.rows = rows
        weights.cols = cols
        weights.sum = 0
        weights.w = [0] * (rows * cols)
        if (
            rows > 0
            and cols > 0
            and len(self.windows) > 0
            and self.statsRegion.height >= rows
            and self.statsRegion.width >= cols
        ):
            maxCellWeight = 46080 // (MAX_WINDOWS * rows * cols)
            cellH = self.statsRegion.height // rows
            cellW = self.statsRegion.width // cols
            cellA = cellH * cellW
            for w in self.windows:
                for r in range(rows):
                    y0 = max(self.statsRegion.y + int(cellH * r), w.y)
                    y1 = min(self.statsRegion.y + int(cellH * (r + 1)), w.y + int(w.height))
                    if y0 >= y1:
                        continue
                    y1 -= y0
                    for c in range(cols):
                        x0 = max(self.statsRegion.x + int(cellW * c), w.x)
                        x1 = min(self.statsRegion.x + int(cellW * (c + 1)), w.x + int(w.width))
                        if x0 >= x1:
                            continue
                        a: float = y1 * (x1 - x0)
                        a = (maxCellWeight * a + cellA - 1) / cellA
                        weights.w[r * cols + c] += a
                        weights.sum += a
        if weights.sum == 0:
            for r in range(rows // 3, rows - rows // 3):
                for c in range(cols // 4, cols - cols // 4):
                    weights.w[r * cols + c] = 1
                    weights.sum += 1

    def __invalidate_weights(self) -> None:
        self.phaseWeights.sum = 0
        self.contrastWeights.sum = 0

    def __get_phase(self, regions: PdafRegions) -> Tuple[float, float, bool]:
        if regions.height != self.phaseWeights.rows or regions.width != self.phaseWeights.cols or self.phaseWeights.sum == 0:
            self.__compute_weights(self.phaseWeights, regions.height, regions.width)

        sumWc = 0
        sumWcp = 0
        for i in range(regions.numRegions):
            w = self.phaseWeights.w[i]
            if w != 0:
                data = regions.pdaf_grid[i // PDAF_STATS_COLS][i % PDAF_STATS_COLS]
                c = data.conf
                if c >= self.cfg.confThresh:
                    if c > self.cfg.confClip:
                        c = self.cfg.confClip
                    c -= self.cfg.confThresh >> 2
                    sumWc += int(w * c)
                    c -= self.cfg.confThresh >> 2
                    sumWcp += int(w * c) * int(data.phase)

        if 0 < self.phaseWeights.sum and self.phaseWeights.sum <= sumWc:
            phase = sumWcp / sumWc
            conf = sumWc / self.phaseWeights.sum
            return phase, conf, True
        else:
            return 0, 0, False

    def __prepare(self, regions: Optional[PdafRegions]) -> None:
        if self.scan_state == ScanState.Trigger:
            self.__startAF()

        if self.initted:
            phase = 0.0
            conf = 0.0
            if regions is not None:
                phase, conf, _ = self.__get_phase(regions)
            self.__doAF(self.prev_contrast, phase, conf)
            self.__update_lens_position()

        if self.pause_flag:
            self.afstatus.pauseState = AfPauseState.Paused if self.scan_state == ScanState.Idle else AfPauseState.Pausing
        else:
            self.afstatus.pauseState = AfPauseState.Running

        if self.afmode == AfMode.AfModeAuto and self.scan_state != ScanState.Idle:
            self.afstatus.state = AfState.Scanning
        else:
            self.afstatus.state = self.report_state
        lensSetting = self.cfg.map.eval(self.fsmooth)
        self.afstatus.lensSetting = int(lensSetting) if self.initted and lensSetting is not None else None
        try:
            if self.afstatus.lensSetting is not None:
                self.callback_fn(self.afstatus.lensSetting)
        except Exception:
            # Occasionally, the execution may fail due to the current state of the device.
            raise RuntimeWarning("Failed to execute the callback function")

    def __get_contrast(self, stats: bcm2835_isp_stats) -> float:
        if self.contrastWeights.sum == 0:
            self.__compute_weights(self.contrastWeights, CONTRAST_STATS_ROWS, CONTRAST_STATS_COLS)
        sumWc = 0
        for i in range(CONTRAST_STATS_ROWS * CONTRAST_STATS_COLS):
            # https://github.com/raspberrypi/libcamera/blob/v0.3.2%2Brpt20240927/src/ipa/rpi/vc4/vc4.cpp#L245
            val = stats.focus_stats[i].contrast_val[1][1] / 1000
            sumWc += self.contrastWeights.w[i] * val
        return sumWc / self.contrastWeights.sum if self.contrastWeights.sum > 0 else 0.0

    def set_unicam_config(
        self,
        sensor_config: Dict,
        unicam_width: int,
        sensor_size: Tuple[int, int],
        post_crop_size: Tuple[int, int, int, int],
        camera_mode_scale: Tuple[float, float],
        camera_mode_crop_size: Tuple[int, int],
        callback_fn: Callable,
    ) -> None:
        """
        This method is called in UnicamIspCapture.__init__().
        """
        if "rpi.af" not in sensor_config:
            raise RuntimeError("rpi.af is not found in sensor config")
        self.cfg = CfgParams(sensor_config["rpi.af"])
        self.line_width = unicam_width
        self.bytes_per_line = (self.line_width * self.bitdepth) >> 3
        self.callback_fn = callback_fn
        self.statsRegion = Rectangle(
            camera_mode_crop_size[0],
            camera_mode_crop_size[1],
            sensor_size[0] - 2 * camera_mode_crop_size[0],
            sensor_size[1] - 2 * camera_mode_crop_size[0],
        )
        # for window coordinate decode
        self.post_crop_size = post_crop_size
        self.camera_mode_scale = camera_mode_scale
        self.camera_mode_crop = camera_mode_crop_size

    def set_focus_windows(self, bboxes: List[List[float]]) -> None:
        """
        Configures the focus windows.
        bboxes: [[x1, y1, x2, y2], ...] defined in capture image coordinate system.
        """
        self.windows.clear()
        for bbox in bboxes[:MAX_WINDOWS]:
            x1, y1, x2, y2 = bbox
            self.windows.append(
                # Transform capture image coordinates to the sensor's coordinate system
                Rectangle(
                    (x1 + self.post_crop_size[0]) * self.camera_mode_scale[0] + self.camera_mode_crop[0],
                    (y1 + self.post_crop_size[1]) * self.camera_mode_scale[1] + self.camera_mode_crop[1],
                    (x2 - x1) * self.camera_mode_scale[0],
                    (y2 - y1) * self.camera_mode_scale[1],
                )
            )
        self.__invalidate_weights()

    def set_focus_value(self, value: int) -> None:
        if self.afmode == AfMode.AfModeManual:
            ftarget = self.cfg.map.inverse_eval(value)
            if ftarget is not None:
                self.afstatus.lensSetting = value
                self.ftarget = ftarget
                self.__update_lens_position()
                self.callback_fn(value)

    def get_focus_value(self) -> Optional[int]:
        """
        Return latest focus value.
        """
        return self.afstatus.lensSetting

    def get_focus_stats(self) -> AfStatus:
        """
        Return latest focus status.
        """
        return self.afstatus

    def trigger_scan(self) -> None:
        if self.afmode == AfMode.AfModeAuto and self.scan_state == ScanState.Idle:
            self.scan_state = ScanState.Trigger

    def process_contrast_metadata(self, stats: bcm2835_isp_stats) -> None:
        """
        Extract contrast information from the metadata received from the ISP device and store it for later use.
        """
        self.prev_contrast = self.__get_contrast(stats)

    @abstractmethod
    def parse_pdaf(self, meta_buffer: VideoBuffer) -> PdafRegions:
        raise NotImplementedError()

    def parse_pdaf_and_update_focus(self, meta_buffer: VideoBuffer) -> None:
        """
        Parse PDAF data from embedded metadata from the unicam meta device.
        Calculate the optimal lens position based on the analyzed data, and execute the callback function.
        """
        if meta_buffer.buf.bytesused < self.bytes_per_line * 3:
            raise RuntimeError("Insufficient metadata size")
        pdaf = self.parse_pdaf(meta_buffer)
        self.__prepare(pdaf)


class AutoFocuserIMX708(AutoFocuserBase):
    def __init__(
        self,
        afrange: AfRange = AfRange.AfRangeNormal,
        afspeed: AfSpeed = AfSpeed.AfSpeedNormal,
        afmode: AfMode = AfMode.AfModeContinuous,
    ) -> None:
        super().__init__(afrange, afspeed, afmode)

    def parse_pdaf(self, meta_buffer: VideoBuffer) -> PdafRegions:
        # https://github.com/raspberrypi/libcamera/blob/v0.3.2%2Brpt20240927/src/ipa/rpi/cam_helper/cam_helper_imx708.cpp#L267-L293
        bpp = 10  # hardcode
        pdaf = PdafRegions(height=PDAF_STATS_ROWS, width=PDAF_STATS_COLS)
        ptr = meta_buffer.mapped_buf[2 * self.bytes_per_line : 3 * self.bytes_per_line]
        step = bpp >> 1

        if bpp < 10 or bpp > 14 or self.bytes_per_line < 194 * step or ptr[0] != 0 or ptr[1] >= 0x40:
            raise RuntimeError("PDAF data in unsupported format")

        offset = 2 * step
        for i in range(PDAF_STATS_ROWS):
            for j in range(PDAF_STATS_COLS):
                conf = (ptr[offset] << 3) | (ptr[offset + 1] >> 5)
                phase = 0
                if conf != 0:
                    phase = (((ptr[offset + 1] & 0x0F) - (ptr[offset + 1] & 0x10)) << 6) | (ptr[offset + 2] >> 2)
                pdaf.pdaf_grid[i][j] = PdafData(conf, phase)
                offset += step
        return pdaf



"""
LibcameraCapture module.

Note:
    This module is only available in ActcastOS4 or later.
"""
import mmap
import selectors
from typing import List, Optional, Tuple

import libcamera as libcam
from actfw_core.capture import Frame
from actfw_core.system import EnvironmentVariableNotSet, get_actcast_firmware_type
from actfw_core.task import Producer
from actfw_core.util.pad import _PadBase, _PadDiscardingOld


class CameraConfigurationInvalidError(Exception):
    _config: libcam.CameraConfiguration
    _msg: str

    def __init__(self, config: libcam.CameraConfiguration, msg: str = "Invalid CameraConfiguration"):
        super().__init__(msg)
        self._config = config
        self._msg = msg

    def __str__(self) -> str:
        return f"{self._msg}: {self._config}"


class CameraConfigureError(Exception):
    _errno: int
    _msg: str

    def __init__(self, errno: int, msg: str = "Camera configure failed"):
        super().__init__(msg)
        self._errno = errno
        self._msg = msg

    def __str__(self) -> str:
        return f"{self._msg}: {self._errno}"


class FrameBufferAllocateError(Exception):
    _errno: int
    _msg: str

    def __init__(self, errno: int, msg: str = "FrameBuffer allocation error"):
        super().__init__(msg)
        self._errno = errno
        self._msg = msg

    def __str__(self) -> str:
        return f"{self._msg}: {self._errno}"


class CameraStartError(Exception):
    _errno: int
    _msg: str

    def __init__(self, errno: int, msg: str = "CameraStart error"):
        super().__init__(msg)
        self._errno = errno
        self._msg = msg

    def __str__(self) -> str:
        return f"{self._msg}: {self._errno}"


class QueueRequestError(Exception):
    _errno: int
    _msg: str

    def __init__(self, errno: int, msg: str = "QueueRequest error"):
        super().__init__(msg)
        self._errno = errno
        self._msg = msg

    def __str__(self) -> str:
        return f"{self._msg}: {self._errno}"


class CaptureTimeoutError(Exception):
    _timeout: int
    _msg: str

    def __init__(self, timeout: int, msg: str = "Capture timeout"):
        super().__init__(msg)
        self._timeout = timeout
        self._msg = msg

    def __str__(self) -> str:
        return f"{self._msg}: {self._timeout}"


class LibcameraCapture(Producer[Frame[bytes]]):
    _cm: libcam.CameraManager
    _size: Tuple[int, int]
    _pixel_format: libcam.PixelFormat
    _camera: libcam.Camera
    _requests: Optional[List[libcam.Request]]
    _camera_config: libcam.CameraConfiguration
    _framerate: int

    def __init__(
        self,
        size: Tuple[int, int],
        pixel_format: libcam.PixelFormat,
        camera_index: int = 0,
        orientation: libcam.Orientation = libcam.Orientation.Rotate0,
        framerate: int = 30,
    ) -> None:
        """
        Initialization method for the LibcameraCapture class.

        Args:
            size: The size of the captured image (width, height).
            pixel_format: The pixel format. Only RGB888 or BGR888 are supported.
                cf. https://libcamera.org/api-html/classlibcamera_1_1PixelFormat.html
            camera_index: The index of the camera to use. Defaults to 0.
            orientation: The orientation of the camera. Defaults to Rotate0.
                cf. https://libcamera.org/api-html/namespacelibcamera.html#a80ea01625b93ecfe879249ac60c79384.
            framerate: The framerate (fps). If not specified, the default setting of the libcamera is used.

        Note:
            As for pixel_format, if RGB888 is specified, BGR888 is actually obtained,
                and if BGR888 is specified, RGB888 is actually obtained.

        Raises:
            AssertionError: Raised if an unsupported pixel format is specified.
            CameraConfigurationInvalidError: Raised if the camera configuration is invalid.
            CameraConfigureError: Raised if the camera configuration fails.
        """
        try:
            firmware_type = get_actcast_firmware_type()
            if firmware_type != "raspberrypi-bookworm":
                raise RuntimeError("LibcameraCapture is only available in ActcastOS4 or later.")
        except EnvironmentVariableNotSet:
            # No error for when running on Raspberry Pi OS
            pass
        assert pixel_format == libcam.PixelFormat("RGB888") or pixel_format == libcam.PixelFormat(
            "BGR888"
        ), "Only RGB888 or BGR888 are supported"
        super().__init__()
        self._cm = libcam.CameraManager.singleton()
        self._size = size
        self._pixel_format = pixel_format
        self._framerate = framerate
        self._camera = self._cm.cameras[camera_index]
        self._camera.acquire()
        self._camera_config = self._camera.generate_configuration([libcam.StreamRole.Viewfinder])
        self._camera_config.orientation = orientation
        stream_config: libcam.StreamConfiguration = self._camera_config.at(0)
        stream_config.size = libcam.Size(*self._size)
        stream_config.pixel_format = self._pixel_format
        res = self._camera_config.validate()
        if res == libcam.CameraConfiguration.Status.Invalid:
            raise CameraConfigurationInvalidError(self._camera_config)
        res = self._camera.configure(self._camera_config)
        if res is not None and res < 0:
            raise CameraConfigureError(res)

    def cameras(self) -> List[libcam.Camera]:
        return self._cm.cameras  # type: ignore

    def capture_size(self) -> Tuple[int, int]:
        stream_config: libcam.StreamConfiguration = self._camera_config.at(0)
        return (stream_config.size.width, stream_config.size.height)

    def _handle_camera_event(self) -> None:
        reqs = self._cm.get_ready_requests()
        for req in reqs:
            buffers = req.buffers
            assert len(buffers) == 1
            stream, frame_buffer = next(iter(buffers.items()))
            assert len(frame_buffer.planes) == 1
            plane = next(iter(frame_buffer.planes))

            with mmap.mmap(plane.fd, plane.length, offset=plane.offset) as mm:
                dst = mm[:]

            frame = Frame(dst)
            self._outlet(frame)

            req.reuse()
            self._camera.queue_request(req)

    def run(self) -> None:
        requests = []
        try:
            allocator = libcam.FrameBufferAllocator(self._camera)
            stream = self._camera_config.at(0).stream
            res = allocator.allocate(stream)
            if res < 0:
                raise FrameBufferAllocateError(res)

            buffers = allocator.buffers(stream)
            for buffer in buffers:
                request = self._camera.create_request()
                res = request.add_buffer(stream, buffer)
                assert res is None or res == 0
                requests.append(request)

            controls = {}
            frame_duration_limit = int(1_000_000 / self._framerate)
            controls[libcam.controls.FrameDurationLimits] = (frame_duration_limit, frame_duration_limit)
            res = self._camera.start(controls=controls)
            if res is not None and res < 0:
                raise CameraStartError(res)

            for request in requests:
                res = self._camera.queue_request(request)
                if res is not None and res < 0:
                    raise QueueRequestError(res)

            sel = selectors.DefaultSelector()
            sel.register(
                self._cm.event_fd,
                selectors.EVENT_READ,
                lambda: self._handle_camera_event(),
            )
            timeout = 1
            while self._is_running():
                events = sel.select(timeout)
                if events == []:
                    raise CaptureTimeoutError(timeout)

                for key, _mask in events:
                    callback = key.data
                    callback()
        finally:
            self._camera.stop()
            self._camera.release()

    def _new_pad(self) -> _PadBase[Frame[bytes]]:
        return _PadDiscardingOld()



import http.server
import io
import socketserver
import threading
from typing import Any, Generic, Optional, TypeVar

from PIL.Image import Image as PIL_Image

from .task import Isolated

PORT = 5100

T = TypeVar("T")


class _ObservableValue(Generic[T]):
    def __init__(self) -> None:
        self.value: Optional[T] = None
        self.condition = threading.Condition()

    def wait_new_value(self) -> T:
        with self.condition:
            self.condition.wait()
            if self.value is None:
                raise RuntimeError("No value has been set")
            return self.value

    def set(self, new_value: T) -> None:
        with self.condition:
            self.value = new_value
            self.condition.notify_all()


class _LocalVideoStreamHandler(http.server.BaseHTTPRequestHandler):
    def __init__(
        self,
        image: _ObservableValue[PIL_Image],
        quality: int,
        *args: Any,
    ) -> None:
        self.image = image
        self.quality = quality
        super().__init__(*args)

    def log_message(self, format: str, *args: Any) -> None:
        """Suppress log messages"""
        pass

    def do_GET(self) -> None:
        self.send_response(200)
        self.send_header("Age", str(0))
        self.send_header("Cache-Control", "no-cache, private")
        self.send_header("Pragma", "no-cache")
        self.send_header("Content-Type", "multipart/x-mixed-replace; boundary=FRAME")
        self.end_headers()
        try:
            while True:
                try:
                    frame = self.image.wait_new_value()
                except Exception:
                    continue
                else:
                    jpgimg = io.BytesIO()
                    frame.save(
                        jpgimg,
                        format="JPEG",
                        quality=self.quality,
                    )
                    self.wfile.write(b"--FRAME\r\n")
                    self.wfile.write(b"Content-Type: image/jpeg\r\n\r\n")
                    self.wfile.write(jpgimg.getvalue())
                    self.wfile.write(b"\r\n")
        except Exception:
            pass


class _LocalVideoStreamServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    allow_reuse_address = True
    daemon_threads = True


class LocalVideoServer(Isolated):
    image: _ObservableValue[PIL_Image]
    server: _LocalVideoStreamServer

    """Local Video Server

    This server provides a local video stream on port 5100.

    """

    def __init__(
        self,
        quality: int = 75,  # 75 is the default value of PIL JPEG quality
    ) -> None:
        """

        Initialize a LocalVideoServer instance.

        Args:
            quality (int, optional):
                The image quality setting. Acceptable values range from 0 to 100.

                - 0: Lowest quality
                - 95: Highest quality
                - 100: JPEG lossless (no compression)

                The `quality` parameter corresponds to the `quality` parameter in PIL's `Image.save` method.
                Defaults to `75`.

        """
        super().__init__()
        self.image = _ObservableValue()

        def handler(*args: Any) -> _LocalVideoStreamHandler:
            return _LocalVideoStreamHandler(self.image, quality, *args)

        self.server = _LocalVideoStreamServer(("", PORT), handler)

    def update_image(self, image: PIL_Image) -> None:
        """

        Update the video image.

        Args:
            image (:class:`~PIL.Image`): image

        """

        try:
            self.image.set(image.copy())
        except Exception:
            pass

    def run(self) -> None:
        self.server.serve_forever()

    def stop(self) -> None:
        super().stop()
        self.server.shutdown()
        self.server.server_close()
